"""Perform JV hysteresis simulations"""
######### Package Imports #########################################################################

import os,sys
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.integrate import trapezoid
# import pySIMsalabim
## Import pySIMsalabim, if not successful, add the parent directory to the system path
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.utils import general as utils_gen
from pySIMsalabim.plots import plot_functions as utils_plot
from pySIMsalabim.utils.utils import update_cmd_pars

######### Function Definitions ####################################################################

def build_tVG_arrays(Vmin,Vmax,scan_speed,direction,steps,G_frac):
    """Build the Arrays for time, voltage and Generation rate for a hysteresis experiment.

    Parameters
    ----------
    Vmin : float 
        minimum voltage
    Vmax : float
        maximum voltage
    scan_speed : float
        Voltage scan speed [V/s]
    direction : integer
        Perform a Vmin-Vmax-Vmin (1) or Vmax-Vmin-Vmax scan (-1)
    steps : integer
        Number of time steps
    G_frac : float
        Device Parameter | Fractional Generation rate

    Returns
    -------
    np.array
        Array of time points
    np.array
        Array of voltages
    np.array
        Array of generation rates
    """    
    # Determine max time point
    tmax = abs((Vmax - Vmin)/scan_speed)
    V, G = [], []
    Vmin_ = Vmin
    Vmax_ = Vmax

    if direction == -1:
        Vmin = Vmax_
        Vmax = Vmin_

    t_min_to_max = np.linspace(0,tmax,int(steps/2))
    t_max_to_min = np.linspace(tmax,2*tmax,int(steps/2))
    t_max_to_min = np.delete(t_max_to_min,[0]) # remove double entry
    t = np.append(t_min_to_max,t_max_to_min)

    for i in t:
        if i < tmax:
            # First  voltage sweep
            V.append(direction*scan_speed*i + Vmin)
        else: 
            # Second voltage sweep
            V.append(-direction*scan_speed*(i-tmax) + Vmax)
        # Append the generation rate
        G.append(G_frac)
    # convert to numpy arrays
    V, G = np.asarray(V), np.asarray(G)
    return t,V,G

def build_tVG_arrays_log(Vmin,Vmax,Vacc,scan_speed,direction,steps,G_frac):
    """Build the Arrays for time, voltage and Generation rate for a hysteresis experiment with an exponential voltage sweep.

    Parameters
    ----------
    Vmin : float 
        minimum voltage
    Vmax : float
        maximum voltage
    scan_speed : float
        Voltage scan speed [V/s]
    direction : integer
        Perform a Vmin-Vmax-Vmin (1) or Vmax-Vmin-Vmax scan (-1)
    steps : integer
        Number of time steps
    G_frac : float
        Device Parameter | Fractional Generation rate
    Vminexpo : float
        Voltage at which the exponential voltage steps start if Vmin or Vmax is 0

    Returns
    -------
    np.array
        Array of time points
    np.array
        Array of voltages
    np.array
        Array of generation rates
    """ 
    if Vacc >= Vmin and Vacc <= Vmax:
        raise ValueError('Vacc must not be between Vmin and Vmax')
    
    d = Vacc-Vmax
    V_min_to_max = [ Vacc - d * np.exp((1 - i/(steps/2 - 1)) * np.log((Vacc - Vmin)/d)) for i in range(steps//2) ]
    V_min_to_max = np.array(V_min_to_max)
    if Vmin == 0:
        # find idx of the min of Vmin_to_max and set it to 0
        min_idx = np.argmin(V_min_to_max)
        V_min_to_max[min_idx] = 0
    elif Vmax == 0:
        # find idx of the max of Vmin_to_max and set it to 0
        max_idx = np.argmax(V_min_to_max)
        V_min_to_max[max_idx] = 0

    V_max_to_min = V_min_to_max[::-1]
    if direction == 1:
        # forward -> backward
        V_max_to_min = np.delete(V_max_to_min,[0])# remove double entry
        V = np.append(V_min_to_max,V_max_to_min)
    elif direction == -1:
        # backward -> forward
        V_min_to_max = np.delete(V_min_to_max,[0])# remove double entry
        V = np.append(V_max_to_min,V_min_to_max)
    G = G_frac * np.ones(len(V))
    # calculate the time array based on the voltage array and the scan speed
    t = np.zeros(len(V))
    for i in range(1,len(V)):
        t[i] = t[i-1] + abs((V[i]-V[i-1])/scan_speed)

    return t,V,G

def create_tVG_hysteresis(session_path, Vmin, Vmax, scan_speed, direction, steps, G_frac, tVG_name, Vacc=None,Vdist=1,**kwargs):
    """Create a tVG file for hysteresis experiments. 

    Parameters
    ----------
    session_path : string
        working directory for zimt
    Vmin : float 
        Left voltage boundary
    Vmax : float
        Right voltage boundary
    scan_speed : float
        Voltage scan speed [V/s]
    direction : integer
        Perform a Vmin-Vmax-Vmin (1) or Vmax-Vmin-Vmax scan (-1)
    steps : integer
        Number of time steps
    G_frac : float
        Device Parameter | Fractional Generation rate
    Vacc : float, optional
        Point of accumulation of row of V's, note: Vacc should be slightly larger than Vmax or slightly lower than Vmin, only needed if Vdist=2, else ignored, default is None
    Vdist : integer, optional
        Voltage distribution type (1: linear, 2: exponential), default is 1
    tVG_name : string
        Device Parameter | Name of the tVG file
    **kwargs : dict
        Additional keyword arguments


    Returns
    -------
    integer
        Value to indicate the result of the process
    string
        A message to indicate the result of the process
    """
    # kwargs
    
    expo_mode = True if Vdist == 2 else False
    # Vminexpo = kwargs.get('Vminexpo', 1e-2) # Voltage at which the exponential voltage steps start if Vmin or Vmax is 0
    # # check that Vminexpo is positive
    # if Vminexpo <= 0:
    #     msg = 'Vminexpo must be strictly positive'
    #     retval = 1
    #     return retval, msg
    if expo_mode and Vacc is None:
        msg = 'When expo_mode is True, Vacc must be provided'
        retval = 1
        return retval, msg
    
    # check that direction is either 1 or -1
    if direction != 1 and direction != -1:
        msg = 'Incorrect scan direction, choose either 1 for a forward - backward scan or -1 for a backward - forward scan'
        retval = 1
        return retval, msg

    # check that Vmin < Vmax
    if Vmin >= Vmax:
        msg = 'Vmin must be smaller than Vmax'
        retval = 1
        return retval, msg    

    # Create two arrays for both time sweeps
    if expo_mode:
        t,V,G = build_tVG_arrays_log(Vmin,Vmax,Vacc,scan_speed,direction,steps,G_frac)
    else:
        t,V,G = build_tVG_arrays(Vmin,Vmax,scan_speed,direction,steps,G_frac)
        
    # Set the correct header for the tVG file
    tVG_header = ['t','Vext','G_frac']

    # Combine t,V,G arrays into a DataFrame
    tVG = pd.DataFrame(np.stack([t,np.asarray(V),np.asarray(G)]).T,columns=tVG_header)

    # Create tVG file
    tVG.to_csv(os.path.join(session_path,tVG_name),sep=' ',index=False,float_format='%.5e')

    # tVG file is created, msg a success
    msg = 'Success'
    retval = 0
    
    return retval, msg

def plot_hysteresis_JV(session_path, path2file = 'tj.dat'):
    """Plot the hysteresis JV curve

    Parameters
    ----------
    session_path : string
        working directory for zimt
    path2file : string
        Path to the tj file

    Returns
    -------
    Axes
        Axes object for the plot

    """
    # Read the data from tj-file
    data_tj = pd.read_csv(os.path.join(session_path,path2file), sep=r'\s+')
    
    fig, ax = plt.subplots()
    pars = {'Jext' : 'Simulation'} #'$J_{ext}$'}
    par_x = 'Vext'
    xlabel = '$V_{ext}$ [V]'
    ylabel = 'Current density [Am$^{-2}$]'
    xscale = 'linear'
    yscale = 'linear'
    title = 'JV curve'
    plot_type = plt.plot

    ax = utils_plot.plot_result(data_tj, pars, list(pars.keys()), par_x, xlabel, ylabel, xscale, yscale, title, ax, plot_type)
    plt.tight_layout() # make it fit in the window
    
    return ax

def tVG_exp(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin, scan_speed, direction, G_frac, tVG_name):
    """Create a tVG file for hysteresis experiments where Vext is the same as the voltages in the experimental JV file

    Parameters
    ----------
    session_path : string
        working directory for zimt
    expJV_Vmin_Vmax : string
        Name of the file of the Vmin-Vmax JV scan
    expJV_Vmax_Vmin : string
        Name of the file of the Vmax-Vmin JV scan
    scan_speed : float
        Voltage scan speed [V/s]
    direction : integer
        Perform a Vmin-Vmax-Vmin (1) or Vmax-Vmin-Vmax scan (-1)
    G_frac : float
        Fractional Generation rate
    tVG_name : string
        Name of the tVG file

    Returns
    -------
    integer
        Value to indicate the result of the process
    string
        A message to indicate the result of the process
    """
    
    if direction == 1:
        JV_forward, JV_backward = read_Exp_JV(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin)
    elif direction == -1:
        JV_backward, JV_forward = read_Exp_JV(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin)
    else:
        # Illegal value for direction given
        msg = 'Incorrect scan direction, choose either 1 for Vmin-Vmax-Vmin scan or -1 for a Vmax-Vmin-Vmax scan'
        retval = 1
        return retval, msg

    V_forward = JV_forward.Vext
    V_backward = JV_backward.Vext

    # Create the time array
    t=np.empty(len(V_forward) + len(V_backward))
    t[0]=0

    # First half
    for i in range(1,len(V_forward)):
        t[i]= t[i-1] + abs((V_forward[i]-V_forward[i-1])/scan_speed)

    # Turning point
    t[len(V_forward)]=t[len(V_forward)-1] + abs((V_backward[0]-V_forward.iloc[-1])/scan_speed)

    # Second half
    for i in range(len(V_forward)+1,len(V_forward) + len(V_backward)):
        t[i]= t[i-1] + abs((V_backward[i-len(V_forward)]-V_backward[i-len(V_forward)-1])/scan_speed)

    # Voltage array
    V = np.concatenate([V_forward, V_backward])

    # Set the correct header for the tVG file
    tVG_header = ['t','Vext','G_frac']

    G = G_frac * np.ones(len(t))

    # Combine t,V,G arrays into a DataFrama
    tVG = pd.DataFrame(np.stack([t,np.asarray(V),G]).T,columns=tVG_header)

    # Create tVG file
    tVG.to_csv(os.path.join(session_path,tVG_name),sep=' ',index=False,float_format='%.3e')

    # tVG file is created, msg a success
    msg = 'Success'
    retval = 0
    return retval, msg

def read_Exp_JV(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin):
    """Read experimental forward and backward JV files

    Parameters
    ----------
    session_path : string
        working directory for zimt
    expJV_Vmin_Vmax : string
        Name of the file of the forward JV scan
    expJV_Vmax_Vmin : string
        Name of the file of the backward JV scan

    Returns
    -------
    np.array
        Array of current and voltage of experimental JV from Vmin to Vmax
    np.array
        Array of current and voltage of experimental JV from Vmax to Vmin
    """
    
    expJV_min_max = os.path.join(session_path, expJV_Vmin_Vmax)
    expJV_max_min = os.path.join(session_path, expJV_Vmax_Vmin)
    
    # Determine time corresponding to each voltage V_i
    JV_min_max = pd.read_csv(expJV_min_max, sep=r'\s+')
    JV_max_min = pd.read_csv(expJV_max_min, sep=r'\s+')
    
    return JV_min_max, JV_max_min

def concatJVs(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin, direction):
    """Put the experimental forward and backward JV arrays together

    Parameters
    ----------
    session_path : string
        working directory for zimt
    expJV_Vmin_Vmax : string
        Name of the file of the forward JV scan
    expJV_Vmax_Vmin : string
        Name of the file of the backward JV scan
    direction : integer
        Perform a Vmin-Vmax-Vmin (1) or Vmax-Vmin-Vmax scan (-1)
    
    Returns
    -------
    np.array
        Array of current and voltage of experimental JV
    """
    
    if direction == 1:
        JV_forward, JV_backward = read_Exp_JV(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin)
    elif direction == -1:
        JV_backward, JV_forward = read_Exp_JV(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin)
    else:
        # Illegal value for direction given
        print('Incorrect scan direction, choose either 1 for Vmin-Vmax-Vmin scan or -1 for a Vmax-Vmin-Vmax scan')
        sys.exit()
    
    expJV = pd.concat([JV_forward, JV_backward], ignore_index=True)   
    return expJV

def read_tj_file(session_path, tj_file_name='tj.dat'):
    """ Read relevant parameters for admittance of the tj file

    Parameters
    ----------
    session_path : string
        Path of the simulation folder for this session
    data_tj : dataFrame
        Pandas dataFrame containing the tj output file from ZimT

    Returns
    -------
    DataFrame
        Pandas dataFrame of the tj_file containing the time, current density, numerical error in the current density and the photogenerated current density
    """

    data = pd.read_csv(os.path.join(session_path,tj_file_name), sep=r'\s+')

    return data

def Compare_Exp_Sim_JV(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin, rms_mode, direction, tj_file_name='tj.dat'):
    """ Calculate the root-mean-square (rms) error of the simulated data compared to the experimental data. The used formulas are
    described in the Manual (see the variable rms_mode in the section 'Description of simulated device parameters').

    Parameters
    ----------
    session_path : string
        Path of the simulation folder for this session
    expJV_Vmin_Vmax : string
        Name of the file of the forward JV scan
    expJV_Vmax_Vmin : string
        Name of the file of the backward JV scan
    rms_mode : string
        Indicates how the normalised rms error should be calculated: either in linear or logarithmic form
    direction : integer
        Perform a Vmin-Vmax-Vmin (1) or Vmax-Vmin-Vmax scan (-1)
    tj_file_name : dataFrame
        Pandas dataFrame containing the tj output file from ZimT

    Returns
    -------
    Float
        Calculated rms-error
    """
        
    JVExp = concatJVs(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin, direction)
    JVSim = read_tj_file(session_path, tj_file_name)[['t', 'Vext', 'Jext']]
    
    # Make an array of voltages that did not converge in simulation
    V_array_not_in_JVSim = np.setdiff1d(JVExp.Vext, JVSim.Vext)
    
    # Remove voltages of experimental data that did not converge in the simulation
    # As the rms-value cannot be calculated, when the voltages of the simulation and experimental data do not overlap
    if len(V_array_not_in_JVSim) > 0:
        disgardedPoints = True
        indices = []
        
        for i in range(len(V_array_not_in_JVSim)):
            # Find indices where voltages do not overlap for every V_i
            index_array = np.where((JVExp.Vext == V_array_not_in_JVSim[i]))[0]
            
            # Add the to-slice-indices to a list
            for j in range(len(index_array)):
                indices.append(index_array[j])
        
        # Delete the indices and convert JVExp from a numpy array in a DataFrame again
        JVExp = np.delete(JVExp, np.sort(indices), axis=0)
        JVExp = pd.DataFrame(JVExp, columns=['Vext', 'Jext'])
    
    rms = 0
    count = 0
    disgardedPoints = False
    
    # Look for the interval [Jmin,Jmax] in both the simulated and experiment data
    Jmin = min(min(JVExp.Jext), min(JVSim.Jext))
    Jmax = max(max(JVExp.Jext), max(JVSim.Jext))
    
    if rms_mode == 'lin' or 'linear':
        # Calculate the sum of squared residuals
        for i in range(len(JVExp)):
            rms = rms + (JVExp.Jext[i] - JVSim.Jext[i])**2
            count += 1
        
        # Calculate the root mean square error and normalise with respect to the interval [Jmin,Jmax]
        rms = np.sqrt(rms/count)/(Jmax-Jmin)
        
    elif rms_mode == 'log' or 'logarithmic':
        # Calculate the sum of squared residuals
        for i in range(len(JVExp)):
            if JVExp.Jext[i]*JVSim.Jext[i]>=0: # We can only calc rms if both are <> 0 and they have the same sign
                rms = rms + np.log(JVExp.Jext[i]/JVSim.Jext[i])**2
            else:
                disgardedPoints = True
            
        # Calculate the root mean square error and normalise with respect to the interval [Jmin,Jmax]
        rms = np.sqrt(rms/count)/abs(np.log(abs(Jmax/Jmin))) # Note: Jmax > Jmin, of course, but ABS(Jmin) can be larger than ABS(Jmax) so we need to use ABS(LN(...)) to get a positive rms

    if disgardedPoints:
        print('Not all JV points were used in computing the rms-error.')
        print('Delete voltages are: ', V_array_not_in_JVSim)
    
    return rms

def calc_hysteresis_index(session_path, tj_file_name = 'tj.dat', tVG_file_ame='tVG.txt', plot_hyst_index = False):
    """
    Calculate the hysteresis index from the simulated JV curve using the difference area between the forward and backward scan
    
    Parameters
    ----------
    session_path : string
        working directory for zimt
    tj_file_name : string
        Name of the tj file
    tVG_file_ame : string
        Name of the tVG file
    plot_hyst_index : bool
        If True, plot the JV curves with the normalisation and difference area

    Returns
    -------
    float
        Hysteresis index
    """

    sign_dum = 0 # dummy value for sign change in voltage array

    # Read data from files. tj file for the JV curve and tVG file to get all possible voltage steps. 
    # This is needed as not all voltages might be present in the tj file.
    data_tj = pd.read_csv(os.path.join(session_path,tj_file_name), sep=r'\s+')
    data_tVG = pd.read_csv(os.path.join(session_path,tVG_file_ame), sep=r'\s+')

    # Store Vinput, Vext and Jext in arrays
    Vinput = np.array(data_tVG['Vext'])
    Vext = np.array(data_tj['Vext'])
    Jext = np.array(data_tj['Jext'])

    # Find the min and max values of Vext and Jext to span the normalisation area 
    Jmin = min(Jext)
    Jmax = max(Jext)
    Vmin = min(Vext)
    Vmax = max(Vext)

    # As the forward and backward scan are in the same array, 
    # we need to find the point where the JV curve changes direction, i.e. the turning point and split into two seperate arrays
    # Vext,Jext array
    diff = np.diff(Vext)
    sign_change = np.where(np.diff(np.sign(diff)))
    
    if len(sign_change[0]) == 0:
        print('Hysteresis index could not be calculated. No sign change in the voltage array')
        return 0
    elif len(sign_change[0]) ==2:
        # Check if they are consecutive. If so, we might just be missing the 'flip'' point, as it could have not converged, so we need to correct for this.
        if sign_change[0][1] - sign_change[0][0] == 1:
            idx_change = sign_change[0][0] + 2
            sign_dum = 1
        else:
            print('Hysteresis index could not be calculated. Multiple sign changes in the voltage array')
            return 0
    elif len(sign_change[0]) > 2:
        print('Hysteresis index could not be calculated. Multiple sign changes in the voltage array')
        return 0
    else:
        # Need to correct for the index change between Vinput and the sign_change array
        idx_change = sign_change[0][0] + 2

    # split the Vext and Jext arrays into two arrays
    Vext_1 = Vext[:idx_change]
    Jext_1 = Jext[:idx_change]
    Vext_2 = Vext[idx_change:]
    Jext_2 = Jext[idx_change:]

    if sign_dum != 1:
        # append the last value of the first array to the first position of the second array to be able to include the turning point
        Vext_2 = np.insert(Vext_2,0,Vext_1[-1])
        Jext_2 = np.insert(Jext_2,0,Jext_1[-1])

    # Vinput array
    diff = np.diff(Vinput)
    sign_change = np.where(np.diff(np.sign(diff)))

    if len(sign_change[0]) == 0:
        print('Hysteresis index could not be calculated. No sign change in the voltage array')
        return 0
    elif len(sign_change[0]) > 1:
        print('Hysteresis index could not be calculated.. Multiple sign changes in the voltage array')
        return 0
    else:
        # Need to correct for the index change between Vinput and the sign_change array
        sign_change_Vinput = sign_change[0][0] + 2

    # We only need one of the arrays, as we will only use this as a reference
    Vinput_1 = Vinput[:sign_change_Vinput]


    # To calculate the difference area between the two curves, we need to reverse the order of the second arrays to match the order of the applied voltage
    # reverse the second array
    Vext_2 = Vext_2[::-1]
    Jext_2 = Jext_2[::-1]

    # Create empty arrays to store the final V,J values
    Vfinal = []
    Jfinal_1 = []
    Jfinal_2 = []

    # Loop over Vinput and check if the voltage exist in both V arrays
    # If it does, add the current to the corresponding J array, else we cannot use it so we discard it
    for i in range(len(Vinput_1)):
        if (Vinput_1[i] in Vext_1) and (Vinput_1[i] in Vext_2):
            # Get the index of the voltage in the corresponding V arrays
            idx1 = np.where(Vext_1 == Vinput_1[i])[0][0]
            idx2 = np.where(Vext_2 == Vinput_1[i])[0][0] 

            # Append the values to the final arrays
            Vfinal.append(Vext_1[idx1])
            Jfinal_1.append(Jext_1[idx1])
            Jfinal_2.append(Jext_2[idx2])

    # Calculate the hysteresis index
    hysteresis_index_num = abs(trapezoid(np.abs(np.array(Jfinal_1) - np.array(Jfinal_2)),Vfinal)) # Use abs around trapezoid to avoid negative values when Vfinal is flipped
    hysteresis_index_denom = (Jmax - Jmin) * (Vmax - Vmin)
    hysteresis_index = hysteresis_index_num / hysteresis_index_denom

    if plot_hyst_index == True:

        # Plot JV traces together with normalisation and difference area
        fig, ax = plt.subplots()

        # Normalisation area
        ax.vlines(x = Vmin,ymin =  Jmin, ymax = Jmax, color='#919191',linestyles='dashed',label='Normalisation area')
        ax.vlines(x = Vmax,ymin =  Jmin, ymax = Jmax, color='#919191',linestyles='dashed')
        ax.hlines(y = Jmin,xmin =  Vmin, xmax = Vmax, color='#919191',linestyles='dashed')
        ax.hlines(y = Jmax,xmin =  Vmin, xmax = Vmax, color='#919191',linestyles='dashed')

        # Difference area
        ax.fill_between(Vfinal,Jfinal_1,Jfinal_2, color='#b1f997',label='Difference area')
    
        # JV traces
        ax.plot(Vfinal,Jfinal_2,label='Backward', color='r', linewidth=4)
        ax.plot(Vfinal,Jfinal_1,label='Forward', color = 'b', linewidth=4)

        ax.set_xlabel('$V_{ext}$ [V]')
        ax.set_ylabel('$J_{ext}$ [Am$^{-2}$]')

        # Reverse the legend order. Curves should be listed first, then norm/diff area. 
        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles[::-1], labels[::-1],framealpha=1, loc='center left', bbox_to_anchor=(0.07, 0.75))
        plt.tight_layout()

        plt.show()

    return hysteresis_index

def Hysteresis_JV(zimt_device_parameters, session_path, UseExpData, scan_speed, direction, G_frac, tVG_name='tVG.txt', tj_name = 'tj.dat',varFile='none',
                  run_mode=False, Vmin=0.0, Vmax=0.0, steps =0, expJV_Vmin_Vmax='', expJV_Vmax_Vmin='',rms_mode='lin', **kwargs ):
    """Create a tVG file and perform a JV hysteresis experiment.

    Parameters
    ----------
    zimt_device_parameters : string
        name of the zimt device parameters file
    session_path : string
        working directory for zimt
    UseExpData : integer
        If 1, use experimental JV curves. If 0, Use Vmin, Vmax as boundaries
    scan_speed : float
        Voltage scan speed [V/s]
    direction : integer
        Perform a forward-backward (1) or backward-forward scan (-1).
    G_frac : float
        Device Parameter | Fractional generation rate
    run_mode : bool, optional
        Indicate whether the script is in 'web' mode (True) or standalone mode (False). Used to control the console output, by default False  
    tVG_name : string, optional
        Device Parameter | Name of the tVG file, by default 'tVG.txt'
    tj_name : string, optional
        Name of the tj file, by default 'tj.dat'
    varFile : string, optional
        Name of the var file, by default 'none'
    run_mode : bool, optional
        indicate whether the script is in 'web' mode (True) or standalone mode (False). Used to control the console output, by default False
    Vmin : float, optional
        Left voltage boundary, by default 0.0
    Vmax : float, optional
        Right voltage boundary, by default 0.0
    steps : int, optional
        Number of time steps, by default 0
    expJV_Vmin_Vmax : str, optional
        file name of the first expJV curve, by default ''
    expJV_Vmax_Vmin : str, optional
        file name of the second expJV curve, by default ''
    rms_mode : str, optional
        Either 'lin' or 'log' to specify how the rms error is calculated

    Returns
    -------
    CompletedProcess
        Output object of with returncode and console output of the simulation
    string
        Return message to display on the UI, for both success and failed
    dict
        Dictionary containing the special output values of the simulation. In this case, the rms error ('rms') and the hysteresis index ('hyst_index')
    """
    verbose = kwargs.get('verbose', False) # Check if the user wants to see the console output
    UUID = kwargs.get('UUID', '') # Check if the user wants to add a UUID to the tj file name
    cmd_pars = kwargs.get('cmd_pars', None) # Check if the user wants to add additional command line parameters
    Vdist = kwargs.get('Vdist', 1) # Voltage distribution type (1: linear, 2: exponential)
    Vacc = kwargs.get('Vacc', None) # Point of accumulation of row of V's, note: Vacc should be slightly larger than Vmax or slightly lower than Vmin, only needed if Vdist=2, else ignored

    # Check if the user wants to force the use of thread safe mode, necessary for Windows with parallel simulations
    if os.name == 'nt':  
        threadsafe = kwargs.get('threadsafe', True) # Check if the user wants to force the use of threads instead of processes
    else:
        threadsafe = kwargs.get('threadsafe', False) # Check if the user wants to force the use of threads instead of processes

    turnoff_autoTidy = kwargs.get('turnoff_autoTidy', None) # Check if the user wants to turn off the autoTidy function in SIMsalabim
    if turnoff_autoTidy is None: 
        if not threadsafe:
            turnoff_autoTidy = True
        else:
            turnoff_autoTidy = False

    # tVG file generation additional parameters
    # expo_mode = kwargs.get('expo_mode', False) # whether to use exponential time steps
    # Vminexpo = kwargs.get('Vminexpo', 1e-2) # minimum voltage after 0 to start the log steps

    # Update the JV file name with the UUID
    if UUID != '':
        dum_str = f'_{UUID}'
    else:
        dum_str = ''

    # Update the filenames with the UUID
    tj_name = os.path.join(session_path, tj_name)
    tVG_name = os.path.join(session_path, tVG_name)
    if UUID != '':
        tj_file_name_base, tj_file_name_ext = os.path.splitext(tj_name)
        tj_name = tj_file_name_base + dum_str + tj_file_name_ext 
        tVG_name_base, tVG_name_ext = os.path.splitext(tVG_name)
        tVG_name = tVG_name_base + dum_str + tVG_name_ext
        if varFile != 'none':
            var_file_base, var_file_ext = os.path.splitext(varFile)
            varFile = var_file_base + dum_str + var_file_ext
            varFile = os.path.join(session_path, varFile)
    # varFile = 'none' # we don't use a var file for the hysteresis JV simulation

    # Init output_vals rms & hyst_index
    rms = 0.0
    hyst_index = 0.0

    if UseExpData == 1:
        # When fitting to experimental data, create a tVG file where Vext is the same as the voltages in the experimental JV file
        if os.path.exists(os.path.join(session_path, expJV_Vmin_Vmax)) and os.path.exists(os.path.join(session_path, expJV_Vmax_Vmin)):
            result, message = tVG_exp(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin, scan_speed, direction, G_frac, tVG_name)
        else:
            result = 1
            message = 'Experimental JV files not found'
    else:
        result, message = create_tVG_hysteresis(session_path, Vmin, Vmax, scan_speed, direction, steps, G_frac, tVG_name, Vacc=Vacc, Vdist=Vdist)

    if result == 0:
        # tVG file created
        Hysteresis_JV_args = [{'par':'dev_par_file','val':zimt_device_parameters},
                              {'par':'tVGFile','val':tVG_name},
                              {'par':'tJFile','val':tj_name},
                              {'par':'varFile','val':varFile},
                              {'par':'logFile','val':'log'+dum_str+'.txt'}
                              ]
        
        if turnoff_autoTidy:
            Hysteresis_JV_args.append({'par':'autoTidy','val':'0'})

        if cmd_pars is not None:
            Hysteresis_JV_args = update_cmd_pars(Hysteresis_JV_args, cmd_pars)

        if threadsafe:
            result, message = utils_gen.run_simulation_filesafe('zimt', Hysteresis_JV_args, session_path, run_mode, verbose=verbose)
        else:
            result, message = utils_gen.run_simulation('zimt', Hysteresis_JV_args, session_path, run_mode, verbose=verbose)

        if result == 0 or result == 95:
            if UseExpData == 1:
                rms = Compare_Exp_Sim_JV(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin, rms_mode, direction, tj_name)
            hyst_index = calc_hysteresis_index(session_path, tj_name, tVG_name)
        else:
            message = message
        
        result = result

        # Put all the output values in a dictionary to be returned. 
        # By putting it in a dictionary we can add any number of values without breaking return arguments
        output_vals = {'rms': rms, 'hyst_index': hyst_index}
    
    return result, message, output_vals

## Running the function as a standalone script
if __name__ == "__main__":
    ## Manual Hysteresis input parameters. These are overwritten by command line arguments if provided
    scan_speed = 1e1 # V/s
    direction = 1 # Scan direction: 1 for Vmin-Vmax-Vmin, -1 for Vmax-Vmin-Vmax
    G_frac = 1 # amount of suns
    UseExpData = 0 # integer, if 1 read experimental data
    Vmin = 0 # lower voltage boundary
    Vmax = 1.15 # upper voltage boundary
    steps = 200 # Number of voltage steps
    expJV_Vmin_Vmax = 'od05_f.txt' # Forward (direction=1)/Backward (direction=-1) JV scan file
    expJV_Vmax_Vmin = 'od05_b.txt' # Backward (direction=1)/Forward (direction=-1) JV scan file

    # Define folder and file paths
    session_path = os.path.join('../../','SIMsalabim','ZimT')

    zimt_device_parameters = 'simulation_setup.txt'

    tVGFile = 'tVG.txt'
    tJFile = 'tj.dat'
    varFile = 'none'

    # UUID = str(uuid.uuid4()) # Add a UUID to the simulation
    UUID = ''

    # Not user input
    rms_mode = 'log' # lin or log

    ############## Command line arguments  ##############
    ## Notes
    ## - The command line arguments are optional and can be provided in any order
    ## - Each command line argument must be provided in the format -par_name value
    ## - Possible arguments include all SIMsalabim parameters and Hysteresis JV specific parameters as listed before
    ## - Special arguments
    ##   - sp : string
    ##     - The session path, i.e. the working directory for the simulation
    ##   - simsetup : string
    ##     - The name of the zimt simulation setup parameters file
    ##   - UUID : string
    ##     - An UUID to add to the simulation (output)

    cmd_pars_dict = {}
    cmd_pars = []

    # Check if any arguments are provided.
    if len(sys.argv) >= 2:
        # Each arguments should be in the format -par val, i.e. in pairs of two. Skip the first argument as this is the script name
        if not len(sys.argv[1:]) % 2 == 0:
            print('Error in command line parameters. Please provide arguments in the format -par_name value -par_name value ...')
            sys.exit(1)
        else:
            input_list = sys.argv[1:]
            # Loop over the input list and put pairs into a dictionary with the first argument as key and the second argument as value
            for i in range(0,len(input_list),2):
                # Check if the key already exists, if not, add it to the cmd_pars_dict
                if str(input_list[i][1:]) in cmd_pars_dict:
                    print(f'Duplicate parameter found in the command line parameters: {str(input_list[i][1:])}')
                    sys.exit(1)
                else:
                    cmd_pars_dict[str(input_list[i][1:])] = str(input_list[i+1])

    # Check and process specific keys/arguments that are not native SIMsalabim arguments
    # Handle the session_path/sp argument separately, as the other parameters depend on this
    if 'sp' in cmd_pars_dict:
        session_path = cmd_pars_dict['sp']
        # remove from cmd_pars
        cmd_pars_dict.pop('sp')

    # Define mappings for keys and variables
    key_action_map = {
        'zimt_device_parameters': lambda val: {'zimt_device_parameters': val},
        'scan_speed': lambda val: {'scan_speed': float(val)},
        'direction': lambda val: {'direction': int(val)},
        'G_frac': lambda val: {'G_frac': float(val)},
        'UseExpData': lambda val: {'UseExpData': int(val)},
        'Vmin': lambda val: {'Vmin': float(val)},
        'Vmax': lambda val: {'Vmax': float(val)},
        'steps': lambda val: {'steps': int(val)},
        'expJV_Vmin_Vmax': lambda val: {'expJV_Vmin_Vmax': val},
        'expJV_Vmax_Vmin': lambda val: {'expJV_Vmax_Vmin': val},
        'tVGFile': lambda val: {'tVGFile':  val},
        'tJFile': lambda val: {'tJFile': val},
        'varFile': lambda val: {'varFile': val},
        'UUID': lambda val: {'UUID': val},
    }

    for key in list(cmd_pars_dict.keys()):  # Use list to avoid modifying the dictionary while iterating
        if key in key_action_map:
            # Apply the corresponding action
            result = key_action_map[key](cmd_pars_dict[key])
            globals().update(result)  # Dynamically update global variables
            cmd_pars_dict.pop(key)

    # Handle remaining keys in `cmd_pars_dict` and add them to the cmd_pars list
    cmd_pars.extend({'par': key, 'val': value} for key, value in cmd_pars_dict.items())
    
    if UseExpData == 1:
        result, message, output_vals = Hysteresis_JV(zimt_device_parameters, session_path, UseExpData, scan_speed, direction, G_frac, tVG_name = tVGFile, tj_name = tJFile,
                                             run_mode = False, expJV_Vmin_Vmax = expJV_Vmin_Vmax, expJV_Vmax_Vmin = expJV_Vmax_Vmin, rms_mode = rms_mode, cmd_pars = cmd_pars, UUID=UUID)
    else:
        result, message, output_vals = Hysteresis_JV(zimt_device_parameters, session_path, UseExpData, scan_speed, direction, G_frac, tVG_name = tVGFile, tj_name = tJFile, 
                                              run_mode = False, Vmin=Vmin, Vmax=Vmax, steps =steps, rms_mode = rms_mode, cmd_pars = cmd_pars, UUID=UUID)
    
    if result == 0 or result == 95:
        if UseExpData == 1:
            print('Rms-value: ', "{:.5f}".format(round(output_vals['rms'], 5)))

        print(f'hyst-index: {output_vals["hyst_index"]:.3f}')

        ax = plot_hysteresis_JV(session_path, tJFile)
        if UseExpData == 1:
            JVExp = concatJVs(session_path, expJV_Vmin_Vmax, expJV_Vmax_Vmin, direction)
            ax.scatter(JVExp.Vext, JVExp.Jext, label='Experimental', color='r')
        
        ax.legend()
        plt.show()
    else:
        print(message) #'Convergence issues, no plot is printed')



""" Test the hysteresis module of pySIMsalabim """

######### Package Imports #########################################################################
import os, sys, uuid 
import pandas as pd
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.experiments.hysteresis import *

######### Test Functions #########################################################################

def test_run_hysteresis_simu():
    """ Test the run_hysteresis_simu function """
    # Set the path to the simulation setup file
    if os.path.exists('SIMsalabim'):
        cwd = os.path.abspath('.')
    else:
        cwd = os.path.abspath('../..')
    zimt_device_parameters = os.path.join(cwd, 'SIMsalabim','ZimT','simulation_setup.txt')
    session_path = os.path.join(cwd, 'SIMsalabim','ZimT')
    # Set the hysteresis parameters
    scan_speed = 0.1
    cmd_pars = [{'par': 'l2.L', 'val': '500e-9'}]
    direction = -1
    G_frac = 1
    tVG_name = os.path.join(session_path,'tVG.txt')
    Vmin = 0.0
    Vmax = 1.3
    steps = 200
    # Run the hysteresis simulation
    ret, mess, rms = Hysteresis_JV(zimt_device_parameters, session_path, 0, scan_speed, direction, G_frac, tVG_name, run_mode=False, Vmin=Vmin, Vmax=Vmax, steps = steps, expJV_Vmin_Vmax='', expJV_Vmax_Vmin='',rms_mode='lin',threadsafe=False,expo_mode = False, Vminexpo = 5e-3, UUID=str(uuid.uuid4()), cmd_pars=cmd_pars)
    # Clean up the output
    sim.clean_all_output(session_path)
    sim.clean_up_output('hysteresis',session_path)
    sim.delete_folders('tmp',session_path)
    # Check the output
    # print(ret)
    assert ret == 0, 'Hysteresis simulation failed'

def test_run_hysteresis_logstep_simu():
    """ Test the test_run_hysteresis_logstep_simu function """
    # Set the path to the simulation setup file
    if os.path.exists('SIMsalabim'):
        cwd = os.path.abspath('.')
    else:
        cwd = os.path.abspath('../..')
    zimt_device_parameters = os.path.join(cwd, 'SIMsalabim','ZimT','simulation_setup.txt')
    session_path = os.path.join(cwd, 'SIMsalabim','ZimT')
    # Set the hysteresis parameters
    scan_speed = 0.1
    cmd_pars = [{'par': 'l2.L', 'val': '500e-9'}]
    direction = -1
    G_frac = 1
    tVG_name = os.path.join(session_path,'tVG.txt')
    Vmin = 0.0
    Vmax = 1.3
    steps = 200
    Vdist = 2
    Vacc = -1e-2
    # Run the hysteresis simulation
    ret, mess, rms = Hysteresis_JV(zimt_device_parameters, session_path, 0, scan_speed, direction, G_frac, tVG_name, run_mode=False, Vmin=Vmin, Vmax=Vmax, steps = steps, expJV_Vmin_Vmax='', expJV_Vmax_Vmin='',rms_mode='lin',threadsafe=False,expo_mode = False, Vminexpo = 5e-3, UUID=str(uuid.uuid4()), cmd_pars=cmd_pars, Vdist=Vdist, Vacc=Vacc)
    # Clean up the output
    sim.clean_all_output(session_path)
    sim.clean_up_output('hysteresis',session_path)
    sim.delete_folders('tmp',session_path)
    # Check the output
    # print(ret)
    assert ret == 0, 'Hysteresis with logarithmic voltage steps simulation failed'

def test_hysteresis_parallel():
    """ Test the hysteresis_parallel function """
    try:
        if os.path.exists('SIMsalabim'):
            cwd = os.path.abspath('.')
        else:
            cwd = os.path.abspath('../..')
        zimt_device_parameters = os.path.join(cwd, 'SIMsalabim','ZimT','simulation_setup.txt')
        session_path = os.path.join(cwd, 'SIMsalabim','ZimT')

        def run(scan_speed,ID): # function to run the hysteresis simulation in parallel
            cmd_pars = [{'par': 'l2.L', 'val': '500e-9'}]
            direction = -1
            G_frac = 1
            tVG_name = os.path.join(session_path,'tVG.txt')
            Vmin = 0.0
            Vmax = 1.3
            steps = 200
            print('Running')
            ret, mess, rms = Hysteresis_JV(zimt_device_parameters, session_path, 0, scan_speed, direction, G_frac, tVG_name, run_mode=False, Vmin=Vmin, Vmax=Vmax, steps = steps, expJV_Vmin_Vmax='', expJV_Vmax_Vmin='',rms_mode='lin',threadsafe=True,expo_mode = False, Vminexpo = 5e-3, UUID=ID, cmd_pars=cmd_pars)

        scan_speeds = np.logspace(-3,3,7)
        ID_list = [str(uuid.uuid4()) for i in range(len(scan_speeds))]

        Parallel(n_jobs=min(len(scan_speeds),10))(delayed(run)(scan_speed,ID) for scan_speed,ID in zip(scan_speeds,ID_list))

        # print('Plotting')
        # plt.figure()
        # for ID,speed in zip(ID_list,scan_speeds):
        #     data_tj = pd.read_csv(os.path.join(session_path,'tj_'+str(ID)+'.dat'), sep=r'\s+')
        #     plt.plot(data_tj['Vext'],data_tj['Jext'],label=str(speed),marker='o')
        #     plt.legend()
        # plt.show()

        sim.clean_all_output(session_path)
        sim.clean_up_output('hysteresis',session_path)
        sim.delete_folders('tmp',session_path)
        assert True
    except Exception as e:
        print(e)
        assert False, 'Hysteresis parallel simulation failed'


if __name__ == '__main__':
    test_run_hysteresis_simu()
    test_hysteresis_parallel()
    print('All hysteresis tests passed')



"""Functions for processing the device parameters"""
######### Package Imports #########################################################################

import os, shutil, random

######### Function Definitions ####################################################################

def load_device_parameters(session_path, dev_par_file_name, default_path=os.path.join('SIMsalabim', 'ZimT'), reset = False, availLayers = [], run_mode = False):
    """Load the device_parameters file and create a List object. Check if a session specific file already exists. 
    If True, use this one, else return to the default device_parameters

    Parameters
    ----------
    session_path : string
        Folder path of the current simulation session
    dev_par_file_name : string
        Name of the device parameters file
    default_path : string
        Path name where the default/standard device parameters file is located
    reset : boolean
        If True, the default device parameters are copied to the session folder
    availLayers : List
        List with all the available layer files
    run_mode : bool, optional
        indicate whether the script is in 'web' mode (True) or standalone mode (False). Used to control the console output, by default True


    Returns
    -------
    List
        List with nested lists for all parameters in all sections.
    List
        List with all the layers
    """
    dev_par = {}
    if run_mode == True:
        # Check if the session specific device parameter file exists. If not, copy the default file to the session folder.
        if not os.path.isfile(os.path.join(session_path, dev_par_file_name)) or (reset == True):
            shutil.copy(os.path.join(default_path, dev_par_file_name), session_path) # The simulation_setup_file
            file_list = os.listdir(default_path)
            for file in file_list:
                if (file.endswith('_parameters.txt')):
                    shutil.copy(os.path.join(default_path, file), session_path) # The layer files

    # Read the simulation_setup file and store all lines in a list
    with open(os.path.join(session_path, dev_par_file_name), encoding='utf-8') as fp:
        # First check how many layers are defined.
        layersSection = False
        layers = [['par', 'setup',dev_par_file_name,dev_par_file_name]] # Initialize the simulation setup file. This is identified by key 'setup'

        for line in fp:
            # Read all lines from the file
            if line.startswith('**'):
            # Left adjusted comment
                comm_line = line.replace('*', '').strip()
                if ('Layers' in comm_line):  # Found section with the layer files
                    layersSection = True
                else:
                    layersSection = False
            else:
                    # Line is either a parameter or leftover comment.
                par_line = line.split('*')
                if '=' in par_line[0]:  # Line contains a parameter
                    par_split = par_line[0].split('=')
                    par = ['par', par_split[0].strip(), par_split[1].strip(),par_line[1].strip()] # The element with index 2 contains the actual file name!
                    if layersSection: # If the line is in the layer section, it contains the name of a layer file, thus add it to the Layers list
                        layers.append(par) # Add sublist to the layers list 
        fp.close()

    # Read each layer file and append it as a sublist in the main dev_par list
    for layer in layers:
        with open(os.path.join(session_path, layer[2]), encoding='utf-8') as fp:
            dev_par[f'{layer[2]}'] = devpar_read_from_txt(fp)
            fp.close()
    
    if run_mode == True:
        # Now load the layer files that are not in the simulation_setup but have been defined or created before
        devLayerList = []
        for layer in layers:
            devLayerList.append(layer[2])

        extraLayers = []
        for layer in availLayers:
            if layer not in devLayerList:
                extraLayers.append(layer)

        # Read each extra layer file and append it as a sublist in the main dev_par list
        for layer in extraLayers:
            with open(os.path.join(session_path, layer), encoding='utf-8') as fp:
                dev_par[f'{layer}'] = devpar_read_from_txt(fp)
                fp.close()
    return dev_par, layers

def devpar_read_from_txt(fp):
    """Read the opened .txt file line by line and store all in a List.

    Parameters
    ----------
    fp : TextIOWrapper
        filepointer to the opened .txt file.

    Returns
    -------
    List
        List with nested lists for all parameters in all sections.
    """
    index = 0
    # Reserve the first element of the list for the top/header description
    dev_par_object = [['Description']]

    # All possible section headers
    section_list = ['General', 'Layers', 'Contacts', 'Optics', 'Numerical Parameters', 'Voltage range of simulation', 'User interface','Mobilities', 'Interface-layer-to-right', 'Ions', 'Generation and recombination', 'Bulk trapping']
    for line in fp:
        # Read all lines from the file
        if line.startswith('**'):
        # Left adjusted comment
            comm_line = line.replace('*', '').strip()
            if (comm_line in section_list):  # Does the line match a section name
                # New section, add element to the main list
                dev_par_object.append([comm_line])
                index += 1
            else:
                # A left-adjusted comment, add with 'comm' flag to current element
                dev_par_object[index].append(['comm', comm_line])
        elif line.strip() == '':
        # Empty line, ignore and do not add to dev_par_object
            continue
        else:
        # Line is either a parameter or leftover comment.
            par_line = line.split('*')
            if '=' in par_line[0]:  # Line contains a parameter
                par_split = par_line[0].split('=')
                par = ['par', par_split[0].strip(), par_split[1].strip(),par_line[1].strip()]
                dev_par_object[index].append(par)
            else:
                # leftover (*) comment. Add to the description of the last added parameter
                dev_par_object[index][-1][3] = dev_par_object[index][-1][3] + \
                    "*" + par_line[1].strip()
    return dev_par_object

def store_file_names(dev_par, sim_type, dev_par_name, layers, run_mode = False): #
    """Read the relevant file names from the device parameters and store the file name in a session state. 
        This way the correct and relevant files can be retrieved and used in the results. Make a distinction between simss and zimt.

    Parameters
    ----------
    dev_par : List
        Nested List object containing all device parameters
    sim_type : string
        Which type of simulation to run: simss or zimt
    dev_par_name : string
        Name of the device parameter file
    layers : List
        List with all the layers in the device
    run_mode : bool, optional
        indicate whether the script is in 'web' mode (True) or standalone mode (False). Used to control the console output, by default True
    """
    
    # make sure sim_type is simss or zimt
    sim = sim_type.lower()
    if sim not in ['simss', 'zimt']:
        raise ValueError('sim_type must be either simss or zimt')

    # st.session_state['LayersFiles'] = []
    LayersFiles = []
    genProfile = 'none'
    expjv_id = False
    varFile = 'none'
    logFile = 'none'
    expJV = 'none'
    JVFile = 'none'
    scParsFile = 'none'
    tVGFile = 'none'
    tJFile = 'none'
    opticsFiles = []

    # Get the relevant file names from the device parameters
    for section in dev_par[dev_par_name][1:]:
        # Generation profile
        if section[0] == 'Optics':
            for param in section:
                if param[1] == 'genProfile':
                    if param[2] != 'none' and param[2] != 'calc':
                        genProfile = param[2]
                        # st.session_state['genProfile'] = param[2]
                    elif param[2] == 'calc':
                        genProfile = 'calc'
                        # st.session_state['genProfile'] = 'calc'
                    else:
                        genProfile = 'none'
                        # st.session_state['genProfile'] = 'none'
        
        # Files in USer Interface section
        if section[0] == 'User interface':
            for param in section:
                if param[1] == 'varFile':
                    varFile = param[2]
                    # st.session_state['varFile'] = param[2]
                if param[1] == 'logFile':
                    logFile = param[2]
                    # st.session_state['logFile'] = param[2]
                if sim == 'simss':
                    if param[1] == 'useExpData':
                        if param[2] != '1':
                            expjv_id = False
                            expJV = 'none'
                            # st.session_state['expJV'] = 'none'
                        else:
                            expjv_id = True
                    if param[1] == 'expJV' and expjv_id is True:
                        expJV = param[2]
                        # st.session_state['expJV'] = param[2]
                    if param[1] == 'JVFile':
                        JVFile = param[2]
                        # st.session_state['JVFile'] = param[2]
                    if param[1] == 'scParsFile':
                        scParsFile = param[2]
                        # st.session_state['scParsFile'] = param[2]
                if sim == 'zimt':
                    if param[1] == 'tVGFile':
                        tVGFile = param[2]
                        # st.session_state['tVGFile'] = param[2]
                    if param[1] == 'tJFile':
                        tJFile = param[2]
                        # st.session_state['tJFile'] = param[2]
                        
        if section[0] == 'Layers':
            for param in section[1:]:
                LayersFiles.append(param[2])
                # st.session_state['LayersFiles'].append(param[2])

    # When the generation profile has been calculated, store the names of the nk and spectrum files. QQQ process for each layer!
    # if st.session_state['genProfile'] == 'calc':
    if genProfile == 'calc':
        opticsFiles = []
        # st.session_state['opticsFiles'] = []
        specfile = ''
        # Get the spectrum and nk files from the simulation setup
        for section in dev_par[dev_par_name][1:]:
            if section[0] == 'Optics':
                for param in section:
                    if param[1].startswith('nk'):
                        opticsFiles.append(param[2])
                        # st.session_state['opticsFiles'].append(param[2])
                    elif param[1]=='spectrum':
                        specfile = param[2]


    # Go over the layer files for trap files and nk files
    # st.session_state['traps_int'] = []
    # st.session_state['traps_bulk'] = []
    traps_int = []
    traps_bulk = []

    usedFiles= [] 
    for layer in layers:
        if not layer[2] in usedFiles: # We only want to check the layer parameter files that have been used in the simulation
            usedFiles.append(layer[2])

    # Get the nk file for each layer
    for usedFile in usedFiles:
        for section in dev_par[usedFile][1:]:
            # if st.session_state['genProfile'] == 'calc':
            if genProfile == 'calc':
                if section[0] == 'Generation and recombination':
                    for param in section:
                        if param[1].startswith('nk'):
                            opticsFiles.append(param[2])
                            # st.session_state['opticsFiles'].append(param[2])
                
            
    # We need to check every layer files whether files for the trap distribution have been used for interface and/or bulk traps.
    # If present, add file name to list, if not add 'none' to the list. We will process this when preparing the download of the files.
            if section[0] == 'Interface-layer-to-right':
                for param in section:
                    if param[1] == 'intTrapFile':
                        if param[2] != 'none':
                            traps_int.append(param[2])
                            # st.session_state['traps_int'].append(param[2])
                        else:
                            traps_int.append('none')
                            # st.session_state['traps_int'].append('none')

            if section[0] == 'Bulk trapping':
                for param in section:
                    if param[1] == 'bulkTrapFile':
                        if param[2] != 'none':
                            traps_bulk.append(param[2])
                            # st.session_state['traps_bulk'].append(param[2])
                        else:
                            traps_bulk.append('none')
                            # st.session_state['traps_bulk'].append('none')

    # Add the name of the spectrum file to the end of the array
    # if st.session_state['genProfile'] == 'calc':
    if genProfile == 'calc':
        opticsFiles.append(specfile)
        # st.session_state['opticsFiles'].append(specfile)

    if run_mode == True:
        return LayersFiles, opticsFiles, genProfile, traps_int, traps_bulk, expJV, varFile, logFile, JVFile, scParsFile, tVGFile, tJFile      
    else:
        if sim == 'simss':
            return LayersFiles, opticsFiles, traps_int, traps_bulk, expJV, varFile, logFile, JVFile, scParsFile
        else:
            return LayersFiles, opticsFiles, traps_int, traps_bulk, tVGFile, varFile, logFile, tJFile


def devpar_read_from_txt(fp):
    """Read the opened .txt file line by line and store all in a List.

    Parameters
    ----------
    fp : TextIOWrapper
        filepointer to the opened .txt file.

    Returns
    -------
    List
        List with nested lists for all parameters in all sections.
    """
    index = 0
    # Reserve the first element of the list for the top/header description
    dev_par_object = [['Description']]

    # All possible section headers
    section_list = ['General', 'Layers', 'Contacts', 'Optics', 'Numerical Parameters', 'Voltage range of simulation', 'User interface','Mobilities', 'Interface-layer-to-right', 'Ions', 'Generation and recombination', 'Bulk trapping']
    for line in fp:
        # Read all lines from the file
        if line.startswith('**'):
        # Left adjusted comment
            comm_line = line.replace('*', '').strip()
            if (comm_line in section_list):  # Does the line match a section name
                # New section, add element to the main list
                dev_par_object.append([comm_line])
                index += 1
            else:
                # A left-adjusted comment, add with 'comm' flag to current element
                dev_par_object[index].append(['comm', comm_line])
        elif line.strip() == '':
        # Empty line, ignore and do not add to dev_par_object
            continue
        else:
        # Line is either a parameter or leftover comment.
            par_line = line.split('*')
            if '=' in par_line[0]:  # Line contains a parameter
                par_split = par_line[0].split('=')
                par = ['par', par_split[0].strip(), par_split[1].strip(),par_line[1].strip()]
                dev_par_object[index].append(par)
            else:
                # leftover (*) comment. Add to the description of the last added parameter
                dev_par_object[index][-1][3] = dev_par_object[index][-1][3] + \
                    "*" + par_line[1].strip()
    return dev_par_object

def devpar_write_to_txt(dev_par_object):
    """Convert the List object into a single string. Formatted to the device_parameter definition

    Parameters
    ----------
    dev_par_object : List
        List object with all parameters and comments.

    Returns
    -------
    string
        Formatted string for the txt file
    """
    par_file = []  # Initialize List to hold all lines
    lmax = 0  # Max width of 'parameter = value' section, initialise with 0
    section_length_max = 84 # Number of characters in the section title

    # Description and Version
    for item in dev_par_object[0][1:]:
        # First element of the main object contains the top description lines. Skip very first element (Title).
        desc_line = "** " + item[1] + '\n'
        par_file.append(desc_line)

    # Determine max width of the 'parameter = value' section of the txt file to align properly.
    for sect_item in dev_par_object[1:]:
        # Loop over all sections
        for par_item in sect_item[1:]:
            # Loop over all parameters
            if par_item[0] == 'par':
                # Only real parameter entries need to be considered, characterised by the first list element being 'par'
                temp_string = par_item[1] + ' = ' + par_item[2]
                if len(temp_string) > lmax:
                    # Update maxlength if length of 'par = val' combination exceeds it.
                    lmax = len(temp_string)
    # Add 1 to max length to allow for a empty space between 'par=val' and description.
    lmax = lmax + 1

    # Read every entry of the Parameter List object and create a formatted line (string) for it. Append to string List par_file.
    for sect_element in dev_par_object[1:]:
        # Loop over all sections. Exclude the first (Description Title) element.

        ## Section
        # Start with a new line before each section name. Section title must be of format **title************...
        par_file.append('\n')
        sec_title = "**" + sect_element[0]
        sec_title_length = len(sec_title)
        sec_title = sec_title + "*" * \
            (section_length_max-sec_title_length) + '\n'
        par_file.append(sec_title)

        ## Parameters
        for par_element in sect_element:
            #  Loop over all elements in the section list, both parameters ('par') and comments ('comm')
            if par_element[0] == 'comm':
                # Create string for a left-justified comment and append to string List.
                par_line = '** ' + par_element[1] + '\n'
                par_file.append(par_line)
            elif par_element[0] == 'par':
                # Create string for a parameter. Format is par = val
                par_line = par_element[1] + ' = ' + par_element[2]
                par_line_length = len(par_line)
                # The string is filled with blank spaces until the max length is reached
                par_line = par_line + ' '*(lmax - par_line_length)
                # The description can be a multi-line description. The multiple lines are seperated by a '*'
                if '*' in par_element[3]:
                    # MultiLine description. Split it and first append the par=val line as normal
                    temp_desc = par_element[3].split('*')
                    par_line = par_line + '* ' + temp_desc[0] + '\n'
                    par_file.append(par_line)
                    for temp_desc_element in temp_desc[1:]:
                        #  For every extra comment line, fill left part of the line with empty characters and add comment/description as normal.
                        par_line = ' '*lmax + '* ' + temp_desc_element + '\n'
                        par_file.append(par_line)
                else:
                    # Single Line description. Add 'par=val' and comment/description together, seperated by a '*'
                    par_line = par_line + '* ' + par_element[3] + '\n'
                    par_file.append(par_line)

    # Join all individual strings/lines together
    par_file = ''.join(par_file)

    return par_file

def get_inputFile_from_cmd_pars(sim_type, cmd_pars):
    """Get the input file name from the command line parameters except the layer files

    Parameters
    ----------
    sim_type : string
        Which type of simulation to run: simss or zimt
    cmd_pars : List
        List with the command line parameters
    except_layers : bool, optional
        If True, the layer files are excluded from the list, by default True

    Returns
    -------
    string
        The name of the input file
    """
    input_files, newlayers = [], []
    # make sure sim_type is simss or zimt
    sim = sim_type.lower()
    if sim not in ['simss', 'zimt']:
        raise ValueError('sim_type must be either simss or zimt')

    ignore_output_files = ['JVFile', 'scParsFile', 'tJFile', 'varFile', 'logFile']

    # Get the input file name from the command line parameters
    if sim == 'simss':
        for cmd_par in cmd_pars:

            # if not except_layers:
            # if cmd_par['par'].startswith('l') and cmd_par['par'][1:].isdigit(): # layerfile
            #     # input_files.append(cmd_par)
            #     print('1')
            #     newlayers.append(cmd_par)

            if cmd_par['par'].endswith('File') and not cmd_par['par'] in ignore_output_files:
                input_files.append(cmd_par)
            
            if cmd_par['par'] == 'expJV' and cmd_par['val'] != 'none':
                input_files.append(cmd_par)
            
            if cmd_par['par'] == 'genProfile' and (cmd_par['val'] != 'calc' and cmd_par['val'] != 'none'):
                input_files.append(cmd_par)

            # for layer parameters split the cmd_pars[par] after .
            dum_par = cmd_par['par'].split('.')[-1]
            if dum_par.startswith('nk') :
                input_files.append(cmd_par)
            
            if cmd_par['par'] =='spectrum':
                input_files.append(cmd_par)

    else:
        for cmd_par in cmd_pars:
            # if cmd_par['par'].endswith('File') and not cmd_par['par'] in ignore_output_files:
            #     # input_files.append(cmd_par)
            #     newlayers.append(cmd_par)

            if cmd_par['par'].endswith('File') and not cmd_par['par'] in ignore_output_files:
                input_files.append(cmd_par)
            
            if cmd_par['par'] == 'genProfile' and (cmd_par['val'] != 'calc' or cmd_par['val'] != 'none'):
                input_files.append(cmd_par)

            # for layer parameters split the cmd_pars[par] after .
            dum_par = cmd_par['par'].split('.')[-1]
            if dum_par.startswith('nk') :
                input_files.append(cmd_par)
            
            if cmd_par['par'] =='spectrum':
                input_files.append(cmd_par)


    return input_files #, newlayers

def get_inputFile_from_layer(layer, session_path):
    """Get the input file name from the layer parameters

    Parameters
    ----------
    layer : List
        List with the layer parameters
    session_path : string
        Folder path of the current simulation session

    Returns
    -------
    string
        The name of the input file
    """
    # read the layer file
    with open(os.path.join(session_path, layer[2]), encoding='utf-8') as fp:
        layer_par = devpar_read_from_txt(fp)
        fp.close()
    section2update = ['Layers', 'Optics', 'Generation and recombination', 'Interface-layer-to-right', 'Bulk trapping']
    ignore_output_files = ['JVFile', 'scParsFile', 'tJFile', 'varFile', 'logFile']
    input_files = []
    for section in layer_par[layer[2]][1:]:
        if section[0] in section2update:
            for param in section[1:]:
                if param[0] == 'par':
                    if param[1].endswith('File') and not param[1] in ignore_output_files:
                        input_files.append(param)
                    if param[1] == 'expJV':
                        if param[2] != 'none':
                            input_files.append(param)
                    if param[1] == 'genProfile':
                        if param[2] != 'calc' and param[2] != 'none':
                            input_files.append(param)
                    if param[1].startswith('nk'):
                        input_files.append(param)
                    if param[1] == 'spectrum':
                        input_files.append(param)
                               
    return input_files

def make_basename_file_cmd_pars(cmd_pars,except_output_files = True):
    """ Update the command line parameters with the basename of the input files

    Parameters
    ----------
    cmd_pars : List
        List with the command line parameters
    except_output_files : bool, optional
        If True, the output files names are not updated, by default True

    Returns
    -------
    List
        List with the updated command line parameters
    """
    ignore_output_files = ['JVFile', 'scParsFile', 'tJFile', 'varFile', 'logFile']
    for idx, cmd_par in enumerate(cmd_pars):
        if cmd_par['par'] == 'dev_par_file':
            cmd_pars[idx]['val'] = os.path.basename(cmd_par['val'])

        if cmd_par['par'] == 'genProfile':
            if cmd_par['val'] != 'calc' and cmd_par['val'] != 'none':
                cmd_pars[idx]['val'] = os.path.basename(cmd_par['val'])

        if cmd_par['par'] == 'expJV':
            if cmd_par['val'] != 'none':
                cmd_pars[idx]['val'] = os.path.basename(cmd_par['val'])
        
        if cmd_par['par'].endswith('File'):
            if except_output_files:
                if not cmd_par['par'] in ignore_output_files:
                    cmd_pars[idx]['val'] = os.path.basename(cmd_par['val'])
            else:
                cmd_pars[idx]['val'] = os.path.basename(cmd_par['val'])
    
        if cmd_par['par'].startswith('l') and cmd_par['par'][1:].isdigit():
            cmd_pars[idx]['val'] = os.path.basename(cmd_par['val'])

        if cmd_par['par'].startswith('nk'):
            cmd_pars[idx]['val'] = os.path.basename(cmd_par['val'])

        if cmd_par['par'] == 'spectrum':
            cmd_pars[idx]['val'] = os.path.basename(cmd_par['val'])

    return cmd_pars

def make_basename_input_files(filename, updateFile = True):
    """Update the layer file with the basename of the input files

    Parameters
    ----------
    filename : string
        path to the layer file
    updateFile : bool, optional
        if True, the layer file is updated with the new file names, else the updated layer file list is returned, by default True

    Returns
    -------
    List
        List with the updated layer parameters
    """    
    # read the layer file
    with open(filename, encoding='utf-8') as fp:
        layer_par = devpar_read_from_txt(fp)
        fp.close()
    
    section2update = ['Layers', 'Optics', 'Generation and recombination', 'Interface-layer-to-right', 'Bulk trapping']
    ignore_output_files = ['JVFile', 'scParsFile', 'tJFile', 'varFile', 'logFile']
    for section in layer_par:
        if section[0] in section2update:
            for param in section[1:]:
                if param[0] == 'par':
                    if param[1].endswith('File') and not param[1] in ignore_output_files:
                        param[2] = os.path.basename(param[2])
                    if param[1].startswith('nk'):
                        param[2] = os.path.basename(param[2])
                    if param[1] == 'spectrum':
                        param[2] = os.path.basename(param[2])

    # Write the updated layer file
    if updateFile == True:
        with open(filename, 'w', encoding='utf-8') as fp:
            fp.write(devpar_write_to_txt(layer_par))
            fp.close()
    else:
        return layer_par


def get_par_from_dev_par(dev_par, par_name):
    """Get the parameter from the device parameters

    Parameters
    ----------
    dev_par : List
        List with the device parameters
    par_name : string
        Name of the parameter

    Returns
    -------
    List
        List with the parameter
    """
    for section in dev_par:
        for param in section[1:]:
            if param[1] == par_name:
                return param 

def ReadParameterFile(path2file):
    """Get all the parameters from the 'Device_parameters.txt' file
    for SIMsalabim and ZimT
    Parameters
    ----------
    path2file : str
        Path to the 'Device_parameters.txt'

    Returns
    -------
    dict
        Contains the parameters and values from the 'Device_parameters.txt'
    """    
    
    lines = []
    ParFileDic = {}
    with open(path2file) as f:
        lines = f.readlines()

    count = 0
    for line in lines:
        line = line.replace(' ', '')
        if line[0] != '*' and (not line.isspace()):
            equal_idx = line.find('=')
            star_idx = line.find('*')
            # print(line[0:equal_idx] , line[equal_idx+1:star_idx])
            ParFileDic[line[0:equal_idx] ] = line[equal_idx+1:star_idx]
            count += 1
            # print(f'line {count}: {line}')   
    return ParFileDic

# def update_dev_par(sim_type, cmd_pars, dev_par, layers):
#     """Update the device parameters with the command line parameters

#     Parameters
#     ----------
#     sim_type : string
#         Which type of simulation to run: simss or zimt
#     cmd_pars : List
#         List with the command line parameters
#     dev_par : List
#         List with the device parameters

#     Returns
#     -------
#     List
#         List with the updated device parameters
#     """
#     # make sure sim_type is simss or zimt
#     sim = sim_type.lower()
#     if sim not in ['simss', 'zimt']:
#         raise ValueError('sim_type must be either simss or zimt')

#     # Get the input file name from the command line parameters
#     # input_files, newlayers = get_inputFile_from_cmd_pars(sim_type, cmd_pars)

#     # Update the device parameters with the command line parameters
#     for 



""" install GNU parallel"""
# Description:
# ------------
# This script contains a function to download SIMsalabim from GitHub and compile it

######### Package Imports ####################################################################

import os,shutil,sys, getpass

######### Function Definitions ################################################################

# def install_parallel_Linux():
#     """ Install GNU Parallel on Linux
#     """   
#     # check if system is Linux
#     if os.name != 'posix':
#         print('This function is only available on Linux systems')
#         sys.exit()
    
#     # check if apt-get is available
#     if shutil.which("apt-get") is None:
#         print('apt-get is not available on this system')
#         sys.exit()

#     # Install GNU Parallel
#     password = getpass('Sudo password:')
#     # run the command sudo apt-get install parallel and enter the password and print the output
#     os.system('echo {} | sudo -S apt-get install parallel'.format(password))

def install_GNU_parallel_Linux(verbose=True):
    """Download GNU parallel if on Linux and install it if possible (requires sudo)

    Parameters
    ----------
    verbose : bool, optional
        Print the download progress, by default False


    Returns
    -------
    None

    Raises
    ------
    Exception
        If GNU parallel could not be installed

    """ 
    
    if os.name == 'posix': # Check if we are on Linux
        # check if GNU parallel is installed
        if shutil.which('parallel') == None:
            # Download GNU parallel
            # Requires sudo so we ask if the user has sudo rights
            overwrite = None
            overwrite = input(f'A sudo password is required to install GNU parallel, do you have sudo rights? (y/n): ')
            while overwrite.lower() not in ['y','n','yes','no']:
                print('Please enter y or n')
                overwrite = input(f'A sudo password is required to install GNU parallel, do you have sudo rights? (y/n): ')
            if overwrite.lower() == 'n' or overwrite.lower() =='no':
                raise Exception('GNU parallel could not be installed')

            # try first with sudo apt-get install parallel
            try:
                # check if apt-get is available
                if shutil.which("apt-get") is None:
                    print('apt-get is not available on this system')
                    sys.exit()
                # Install GNU Parallel
                password = getpass('Sudo password:')
                # run the command sudo apt-get install parallel and enter the password and print the output
                os.system('echo {} | sudo -S apt-get install parallel'.format(password))
                # os.system('sudo apt-get install parallel')
                if verbose:
                    print('GNU parallel was installed with apt-get install parallel')
            except:
                # if that doesn't work, download GNU parallel from git
                try:
                    password = getpass('Sudo password:')
                    os.system('wget http://ftp.gnu.org/gnu/parallel/parallel-latest.tar.bz2')
                    os.system('echo {} | sudo -S tar -xjf parallel-latest.tar.bz2'.format(password))
                    os.system('cd parallel-*/')
                    os.system('echo {} | sudo -S ./configure && make'.format(password))
                    os.system('echo {} | sudo -S  sudo make install'.format(password))
                    os.system('rm -rf parallel-*/')
                    os.system('rm parallel-latest.tar.bz2')
                    if verbose:
                        print('GNU parallel was downloaded from git and installed')
                except:
                    print('GNU parallel could not be installed')
                    raise Exception('GNU parallel could not be installed')
        else:
            if verbose:
                print('GNU parallel is already installed')
    else:
        if verbose:
            print('GNU parallel is not installed because you are not on Linux')



if __name__ == '__main__':
    
    install_GNU_parallel_Linux(verbose=True)
    

    
             


"""Draw the energy band diagram"""
######### Package Imports #########################################################################

import os
import matplotlib.pyplot as plt
import pySIMsalabim
from pySIMsalabim.utils import device_parameters as utils_dev

######### Function Definitions ####################################################################

def create_width_label(x_left, x_right, x_corr, value, y_min, ax, color):
    """ Create the label to display the width of a layer on the width bar

    Parameters
    ----------
    x_left : float
        Left x position of the layer [m]
    x_right : float
        right x position of the layer [m]
    x_corr : float
        Correction factor to place the label in the middle of the layer
    value : float
        Width of the layer [m]
    y_max : float
        Lowest energy level of the device [eV]
    ax : axes
        Axes object for the plot
    color : string
        Color of the label
    """
    ax.text((x_left+x_right)/2-x_corr, y_min-2.0, round(value*1e9), color=color)


def plot_device_widths(ax, y_min, L, L_original):
    """Plot a width bar below the band energy diagram with the thickness of each layer.

    Parameters
    ----------
    ax : axes
        Axes object for the plot
    y_max : float
        Lowest energy level [eV]
    L : float
        Full width of the device [m]
    L_original : List
        List with the layer widths before scaling
    """
    # Horizontal line below the band diagram 
    ax.hlines(y_min-1.4, 0, sum(L), color='k')

    # Add a small vertical line at each boundary/interface between two layers and add a label with the width of the layer
    for i in range(len(L)):
        ax.vlines(sum(L[:i]), y_min-1.2, y_min-1.6, color='k')
        create_width_label(sum(L[:i]), sum(L[:i+1]), 0.02*sum(L), L_original[i], y_min, ax, 'k')
    
    # Add the vertical line at the right most side 
    ax.vlines(sum(L), y_min-1.2, y_min-1.6, color='k')

    # Add label with unit of [nm] to the end of the horizontal label
    ax.text(1.04*sum(L), y_min-2.0, '[nm]', color='k')


def create_energy_label(x_left, x_right, L, y, band_type, position, ax, vert_pos='top'):
    """Create and place the label for an energy level (in eV) of a layer

    Parameters
    ----------
    x_left : float
        Left x position of the layer [m]
    x_right : float
        right x position of the layer [m]
    L : float
        Full width of the device [m]
    y : float
        Energy of the band [eV]
    band_type : string
        Type of band (CB, VB, Electrode)
    position : float
        Full length of the device
    ax : axes
        Axes object for the plot
    """

    # Offset of the labels to not overlap with the layers
    if vert_pos == 'top':
        offset = 0.11
    else:
        offset = -0.4

    # If the layer covers over 20% of the figure size, move the label to the x middle of the figure. Else align it to the left side of the layer.
    if (x_right - x_left) > 0.2*position:
        ax.text((x_left+x_right)/2 + 0.01*L, y+offset, y)
    elif 'WR' in band_type:
        # In case of the right WF, we must align the text to the right of the label
        ax.text(x_right, y+offset, y, horizontalalignment='right')
    else:
        ax.text(x_left+ 0.01*L, y+offset, y)

def get_param_band_diagram(dev_par, layers, dev_par_name):
    """Create and display the band diagram on the UI based on the relevant parameters from the dict object

    Parameters
    ----------
    dev_par : dict
        Dictionary with all data
    layers : List
        List with all the layers in the device
    dev_par_name : string
        Name of the device parameter file
    """
    msg = '' # Init error message string

    # Init arrays for thicknesses and energy levels
    L = []
    E_c = []
    E_v = []

    # Get the work functions of the electrodes
    for section in dev_par[dev_par_name]:
        if section[0] == 'Contacts':
            for param in section:
                if 'W_L' in param[1]:
                    W_L = -float(param[2])
                elif 'W_R' in param[1]:
                    W_R = -float(param[2])


    # Get the thicknesses and energy levels from the respective layer files
    for layer in layers[1:]:
        for section in dev_par[layer[2]]:
            if section[0] == 'General':
                for param in section:
                    if param[1] in 'L':
                        L.append(float(param[2]))
                    elif param[1] in 'E_c':
                        E_c.append(-float(param[2]))
                    elif param[1] in 'E_v':
                        E_v.append(-float(param[2]))

    
    # Create a figure where the band diagram will be plotted
    # Each element from the array is a layer in the device
    fig, ax = plt.subplots(figsize = (15,5))

    E_high = max(E_v) # To properly place the horizontal width bar
    L_total = sum(L) # Total thickness of the device

    L_real = L.copy() # Create a backup of the widths before adjusting for plotting, used for creating the width labels

    # Create acolor list with 25 standard colors. 
    colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf', '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf', '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd']

    # Create and plot a color block to indicate the bands of a layer
    for i in range(len(L)):
        # If the layer thickness is less than 9% of the full device thickness, rescale to 9% to remain visible.
        if L[i] < 0.09*L_total:
            L[i] = 0.09*L_total

        ax.fill_between([sum(L[:i]), sum(L[:i+1])], [E_c[i], E_c[i]], y2=[E_v[i], E_v[i]], color=colors[i])
        create_energy_label(sum(L[:i]), sum(L[:i+1]),sum(L), E_c[i], 'CB', L_total, ax, 'top')
        create_energy_label(sum(L[:i]), sum(L[:i+1]),sum(L), E_v[i], 'VB', L_total, ax, 'bot')
    
    L_total = sum(L) # update with the corrected values

    # Based on the alignment of W_R/W_L set the proper position of the labels
    if W_L > W_R:
        W_L_pos = 'top'
        W_R_pos = 'bot'
    else:
        W_L_pos = 'bot'
        W_R_pos = 'top'   

    # Left Electrode
    ax.plot([-0.06*L_total, 0], [W_L, W_L], color='k')
    create_energy_label(-0.08*L_total, 0, sum(L), W_L, 'WL', L_total, ax, W_L_pos)

    # Right Electrode
    ax.plot([L_total, L_total+0.06*L_total], [W_R, W_R], color='k')
    create_energy_label(L_total, L_total+0.08*L_total, sum(L), W_R, 'WR', L_total, ax, W_R_pos)

    # Hide the figure axis
    ax.axis('off')

    # Add a horizontal bar to the figure width the layer widths for an arbitray number of layers
    plot_device_widths(ax, E_high, L, L_real)

    if msg != '':
        # Error encountered, return message
        return msg
    else:
        return fig

def plot_band_diagram(dev_par_file_name, session_path ):
    """Plot the band diagram of the device based on the device parameter file and the session path

    Parameters
    ----------
    dev_par_file_name : string
        Name of the device parameter file
    session_path : string
        Path to the session folder
    """
    # Load the device parameter file
    dev_par,layers = pySIMsalabim.utils_dev.load_device_parameters(session_path, dev_par_file_name)

    # Create and display the band diagram
    fig = get_param_band_diagram(dev_par, layers, dev_par_file_name)
    plt.show()
    return fig


"""Useful helper functions for pySIMsalabim"""
######### Package Imports #########################################################################

import numpy as np
from scipy import stats,constants

## Physics constants
q = constants.value(u'elementary charge')
eps_0 = constants.value(u'electric constant')
kb = constants.value(u'Boltzmann constant in eV/K')

######### Function Definitions ####################################################################

def sci_notation(number, sig_fig=2):
    """Make proper scientific notation for graphs

    Parameters
    ----------
    number : float
        Number to put in scientific notation.

    sig_fig : int, optional
        Number of significant digits (Defaults = 2).

    Returns
    -------
    output : str
        String containing the number in scientific notation
    """
    if sig_fig != -1:
        if number == 0:
            output = '0'
        else:
            ret_string = "{0:.{1:d}e}".format(number, sig_fig)
            a,b = ret_string.split("e")
            if int(b) >= 0:
                b = int(b) #removed leading "+" and strips leading zeros too.
                c = ''
            else: 
                b = abs(int(b))
                c = u"\u207B" # superscript minus sign
            SUP = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
            b = str(b).translate(SUP)
            output =a + ' x 10' + c + b
    else:
        if number == 0:
            output = '0'
        else: 
            ret_string = "{0:.{1:d}e}".format(number, 0)
            a,b = ret_string.split("e")
            b = int(b) #removed leading "+" and strips leading zeros too.
            if int(b) >= 0:
                b = int(b) #removed leading "+" and strips leading zeros too.
                c = ''
            else: 
                b = abs(int(b))
                c = u"\u207B" # superscript minus sign
            SUP = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
            #SUB = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")
            b = str(b).translate(SUP)
            output = '10' + c + b    
    return output

def get_Jsc(Volt,Curr):
    """Get the short-circuit current (Jsc) from solar cell JV-curve by interpolating the current at 0 V

    Parameters
    ----------
    Volt : 1-D sequence of floats
        Array containing the voltages.

    Curr : 1-D sequence of floats
        Array containing the current-densities.

    Returns
    -------
    Jsc : float
        Short-circuit current value
    """
    Jsc_dumb = np.interp(0, Volt, Curr)
    return Jsc_dumb

def get_Voc(Volt,Curr):
    """Get the Open-circuit voltage (Voc) from solar cell JV-curve by interpolating the Voltage when the current is 0

    Parameters
    ----------
    Volt : 1-D sequence of floats
        Array containing the voltages.

    Curr : 1-D sequence of floats
        Array containing the current-densities.

    Returns
    -------
    Voc : float
        Open-circuit voltage value
    """
    Voc_dumb = np.interp(0, Curr, Volt)
    return Voc_dumb

def get_FF(Volt,Curr):
    """Get the fill factor (FF) from solar cell JV-curve by calculating the maximum power point

    Parameters
    ----------
    Volt : 1-D sequence of floats
        Array containing the voltages.

    Curr : 1-D sequence of floats
        Array containing the current-densities.

    Returns
    -------
    FF : float
        Fill factor value
    """
    power = []
    Volt_oc = get_Voc(Volt,Curr)
    Curr_sc = get_Jsc(Volt,Curr)
    for i,j in zip(Volt,Curr):
        if (i < Volt_oc and j > Curr_sc):
            power.append(i*j)
    power_max = min(power)
    FF_dumb = power_max/(Volt_oc*Curr_sc)
    return abs(FF_dumb)

def get_PCE(Volt,Curr,suns=1):
    """Get the power conversion efficiency (PCE) from solar cell JV-curve

    Parameters
    ----------
    Volt : 1-D sequence of floats
        Array containing the voltages.

    Curr : 1-D sequence of floats
        Array containing the current-densities.

    Returns
    -------
    PCE : float
        Power conversion efficiency value.
    """
    Voc_dumb = get_Voc(Volt,Curr)
    Jsc_dumb = get_Jsc(Volt, Curr)
    FF_dumb = get_FF(Volt, Curr)
    PCE_dumb = Voc_dumb*Jsc_dumb*FF_dumb/(10*suns) # to get it in % when Jsc is in A/m2 and Voc in V
    return abs(PCE_dumb)

def get_ideality_factor(suns,Vocs,T=295):
    """Returns ideality factor from suns-Voc data linear fit of Voc = (nIF/Vt)*log(suns) + intercept

    Parameters
    ----------
    suns : 1-D sequence of floats
        Array containing the intensity in sun.

    Vocs : 1-D sequence of floats
        Array containing the open-circuit voltages.
    
    T : float optional
        Temperature in Kelvin (Default = 295 K).

    Returns
    -------
    nIF : float
        Ideality factor value.

    intercept : float
        Intercept of the regression line.

    rvalue : float
        Correlation coefficient.

    pvalue : float
        Two-sided p-value for a hypothesis test whose null hypothesis is
        that the slope is zero, using Wald Test with t-distribution of
        the test statistic.

    stderr : float
        Standard error of the estimated gradient.
    """
    Vt = kb*T
    suns = np.log(suns)
    slope_d, intercept_d, r_value_d, p_value_d, std_err_d = stats.linregress(suns,Vocs)
    nIF = slope_d/Vt
    return nIF,intercept_d, r_value_d**2, p_value_d, std_err_d



"""Perform capacitance simulations"""
######### Package Imports #########################################################################

import os, sys
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import math
import scipy.integrate
# import pySIMsalabim
## Import pySIMsalabim, if not successful, add the parent directory to the system path
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.utils import general as utils_gen
from pySIMsalabim.plots import plot_functions as utils_plot
from pySIMsalabim.utils.utils import *
from pySIMsalabim.utils.device_parameters import *

######### Function Definitions ####################################################################

def create_tVG_CV(V_0, V_max, del_V, V_step, G_frac, tVG_name, session_path, freq, ini_timeFactor, timeFactor):
    """Create a tVG file for capacitance experiment. 

    Parameters
    ----------
    V_0 : float
        Initial voltage
    V_max : float
        Maximum voltage
    del_V : float
        Voltage step that is applied directly after t=0
    V_step : float
        Voltage difference, determines at which voltages the capacitance is determined
    G_frac : float
        Fractional light intensity
    tVG_name : string
        Name of the tVG file
    session_path : string
        Path of the simulation folder for this session
    freq : float
        Frequency at which the capacitance-voltage measurement is performed
    ini_timeFactor : float
        Constant defining the size of the initial timestep
    timeFactor : float
        Exponential increase of the timestep, to reduce the amount of timepoints necessary. Use values close to 1.

    Returns
    -------
    string
        A message to indicate the result of the process
    """        

    # Starting line of the tVG file: header + datapoints at time=0. Set the correct header
    tVG_lines = 't Vext G_frac\n'

    # Loop until V reaches V_max, in other words go from Vmin to Vmax with in V_step steps. Add some extra margin on the Vmax to prevent missing the last voltage point due to numerical accuracy.
    while V_0 <= V_max + V_max*1E-5:
        time = 0
        del_t = ini_timeFactor/freq
        tVG_lines += f'{time:.3e} {V_0:.3e} {G_frac:.3e}\n'
        
        # Make the other lines in the tVG file
        while time < 1/freq: #max time: 1/f_min is enough!
            time += del_t
            del_t = del_t * timeFactor
            tVG_lines += f'{time:.3e} {V_0+del_V:.3e} {G_frac:.3e}\n'
    
        V_0 += V_step

    # Write the tVG lines to the tVG file
    with open(os.path.join(session_path,tVG_name), 'w') as file:
        file.write(tVG_lines)

    # tVG file is created, message a success
    msg = 'Success'
    retval = 0

    return retval, msg

def create_tVG_CV_Rseries(Vint, del_V, G_frac, tVG_name, session_path, freq, ini_timeFactor, timeFactor):
    """Create a tVG file for capacitance experiment. 

    Parameters
    ----------
    V_int : float
        Adjusted internal voltage to correct for series resistance later
    del_V : float
        Voltage step that is applied directly after t=0
    G_frac : float
        Fractional light intensity
    tVG_name : string
        Name of the tVG file
    session_path : string
        Path of the simulation folder for this session
    freq : float
        Frequency at which the capacitance-voltage measurement is performed
    ini_timeFactor : float
        Constant defining the size of the initial timestep
    timeFactor : float
        Exponential increase of the timestep, to reduce the amount of timepoints necessary. Use values close to 1.

    Returns
    -------
    string
        A message to indicate the result of the process
    """        

    # Starting line of the tVG file: header + datapoints at time=0. Set the correct header
    tVG_lines = 't Vext G_frac\n'

    # Loop until V reaches V_max, in other words go from Vmin to Vmax with in V_step steps. Add some extra margin on the Vmax to prevent missing the last voltage point due to numerical accuracy.
    for V_0 in Vint:
        time = 0
        del_t = ini_timeFactor/freq
        tVG_lines += f'{time:.3e} {V_0:.3e} {G_frac:.3e}\n'
        
        # Make the other lines in the tVG file
        while time < 1/freq:
            time += del_t
            del_t = del_t * timeFactor
            tVG_lines += f'{time:.3e} {V_0+del_V:.3e} {G_frac:.3e}\n'

    # while V_0 <= V_max + V_max*1E-5:
    #     time = 0
    #     del_t = ini_timeFactor/freq
    #     tVG_lines += f'{time:.3e} {V_0:.3e} {G_frac:.3e}\n'
        
    #     # Make the other lines in the tVG file
    #     while time < 1/freq: #max time: 1/f_min is enough!
    #         time += del_t
    #         del_t = del_t * timeFactor
    #         tVG_lines += f'{time:.3e} {V_0+del_V:.3e} {G_frac:.3e}\n'
    
    #     V_0 += V_step

    # Write the tVG lines to the tVG file
    with open(os.path.join(session_path,tVG_name), 'w') as file:
        file.write(tVG_lines)

    # tVG file is created, message a success
    msg = 'Success'
    retval = 0

    return retval, msg

def create_tVG_SS_CV(V_min, V_max,Vstep, G_frac, tVG_name, session_path):
    """ Creates the tVG file for the steady state simulation with only t 0 and V_0

    Parameters
    ----------
    V_min : float 
        Initial voltage
    V_max : float
        Maximum voltage
    del_V : float
        Voltage step that is applied after t=0
    G_frac : float
        Fractional light intensity
    tVG_name : string
        Name of the tVG file
    session_path : string
        Path of the simulation folder for this session

    Returns
    -------
    string
        A message to indicate the result of the process

    
    """    

    # Starting line of the tVG file: header + datapoints at time=0. Set the correct header
    tVG_lines = 't Vext G_frac\n' #+ f'{0} {V_0} {G_frac:.3e}\n'
    V_0 =V_min
    while V_0 <= V_max + V_max*1E-5:
        time = 0
        tVG_lines += f'{time:.3e} {V_0:.3e} {G_frac:.3e}\n'
        V_0 += Vstep

    # Write the tVG lines to the tVG file
    with open(os.path.join(session_path,tVG_name), 'w') as file:
        file.write(tVG_lines)

    # tVG file is created, message a success
    msg = 'Success'
    retval = 0

    return retval, msg

def calc_capacitance_forOneVoltage(I, errI, time, VStep, freq):
    """Fourier Decomposition formula which computes the capacitance at frequency freq (Hz) and its complex error
    Based on S.E. Laux, IEEE Trans. Electron Dev. 32 (10), 2028 (1985), eq. 5b
    We integrate from 0 to time[imax] at frequency 1/time[imax]

    Parameters
    ----------
    I : np.array
        Array of currents
    errI : np.array
        Numerical error in calculated currents (output of ZimT)
    time : np.array
        Array with all time positions, from 0 to tmax
    VStep : float
        Voltage step
    imax : integer
        Index of the last timestep/first frequency for which the integrals are calculated

    Returns
    -------
    float
        Capacitance at frequency f: C(f)
    float
        Numerical error in calculated capacitance
    """

    imax = len(I)
    Iinf = I[imax-1] # I at infinite time, i.e. the last one we have.
	
    #prepare array for integrants:
    int2 = np.empty(imax)
    int4 = np.empty(imax)
	
    for i in range(imax):
        cosfac = math.cos(2*math.pi*freq*time[i])
        int2[i] = cosfac*(I[i] - Iinf)	
        int4[i] = cosfac*(I[i] + errI[i] - Iinf - errI[imax-1])	

    #Compute the capacitance:
    cap = scipy.integrate.trapezoid(int2, time)/VStep
	
    #and again, but now with the error added to the current:	
    capErr = scipy.integrate.trapezoid(int4, time)/VStep

    # error is the difference between cap and capPlusErr:
    errC = abs( cap - capErr )

    #now return capacitance, its error and the corresponding frequency:	
    return cap, errC

def calc_capacitance(data, del_V, freq):
    """ Calculate the capacitance over the time range
    
    Parameters
    ----------
    data : dataFrame
        Pandas dataFrame containing the time, voltage, current density and numerical error in the current density of the tj_file
    del_V : float
        Voltage step that is applied directly after t=0
    freq : float
        Frequency at which the capacitance-voltage measurement is performed

    Returns
    -------
    np.array
        Array of the capacitance
    np.array
        Array of the capacitance error
    """
       
    idx_time_zero = data.index[data['t'] == 0] # Find all indices where time is equal to 0
    amountOfVoltagePoints = len(idx_time_zero)
    
    # Prepare array for capacitance & its error:
    cap = np.empty(amountOfVoltagePoints)
    errC = np.empty(amountOfVoltagePoints)

    # Get the capacitance for each voltage step by taking the Fourier Transform over the t & I arrays
    for i in range(amountOfVoltagePoints-1):
        start_idx = idx_time_zero[i]
        end_idx = idx_time_zero[i+1]-1
                
        Jext_subarray = np.array( data.loc[start_idx:end_idx, 'Jext'] )
        errJ_subarray = np.array( data.loc[start_idx:end_idx, 'errJ'] )
        time_subarray = np.array( data.loc[start_idx:end_idx, 't'] )
        
        cap[i], errC[i] = calc_capacitance_forOneVoltage(Jext_subarray, errJ_subarray, time_subarray, del_V, freq)
        
    # Handle the last index from t=0 to tmax_n
    Jext_subarray = np.array( data.loc[idx_time_zero[-1]:, 'Jext'] )
    errJ_subarray = np.array( data.loc[idx_time_zero[-1]:, 'errJ'] )
    time_subarray = np.array( data.loc[idx_time_zero[-1]:, 't'] )
    cap[-1], errC[-1] = calc_capacitance_forOneVoltage(Jext_subarray, errJ_subarray, time_subarray, del_V, freq)
    
    return cap, errC

def calc_impedance_limit_time(I, errI, time, VStep, imax):
    """Fourier Decomposition formula which computes the impedance at frequency freq (Hz) and its complex error
    Based on S.E. Laux, IEEE Trans. Electron Dev. 32 (10), 2028 (1985), eq. 5a, 5b
    We integrate from 0 to time[imax] at frequency 1/time[imax]

    Parameters
    ----------
    I : np.array
        Array of currents
    errI : np.array
        Numerical error in calculated currents (output of ZimT)
    time : np.array
        Array with all time positions, from 0 to tmax
    VStep : float
        Voltage step
    imax : integer
        Index of the last timestep/first frequency for which the integrals are calculated

    Returns
    -------
    float
        Frequency belonging to Z(f)
    complex number
        Impedance at frequency f: Z(f)
    complex number
        Numerical error in calculated impedance Z(f)
    """

    freq=1/time[-1] #we obtain the frequency from the time array
    Iinf = I[-1] # I at infinite time, i.e. the last one we have.
    imax = len(I)
    #prepare array for integrants:
    int1 = np.empty(imax)
    int2 = np.empty(imax)
    int3 = np.empty(imax)
    int4 = np.empty(imax)
	
    #now we use only part of the time array:
    timeLim = time
    for i,t in enumerate(time) :
        sinfac = math.sin(2*math.pi*freq*timeLim[i])
        cosfac = math.cos(2*math.pi*freq*timeLim[i])
        int1[i] = sinfac*(I[i] - Iinf)
        int2[i] = cosfac*(I[i] - Iinf)	
        int3[i] = sinfac*(I[i] + errI[i] - Iinf - errI[-1])
        int4[i] = cosfac*(I[i] + errI[i] - Iinf - errI[-1])	

    #now compute the conductance and capacitance:
    cond = (Iinf - I[0] + 2*math.pi*freq*scipy.integrate.trapezoid(int1, timeLim))/VStep
    cap = scipy.integrate.trapezoid(int2, timeLim)/VStep
    #convert to impedance:
    Z = 1/(cond + 2J*math.pi*freq*cap)
	
    #and again, but now with the error added to the current:	
    condErr = (Iinf + errI[-1] - I[0] - errI[0] + 2*math.pi*freq*scipy.integrate.trapezoid(int3, timeLim))/VStep
    capErr = scipy.integrate.trapezoid(int4, timeLim)/VStep
    #convert to impedance:
    Z2 = 1/(condErr + 2J*math.pi*freq*capErr)
    
    #error is the difference between Z and Z2:
    errZ = Z - Z2
    
    #now return complex impedance, its error and the corresponding frequency:	
    return freq, Z, errZ

def calc_impedance_CV(data, del_V, isToPlot,session_path,zimt_device_parameters,Rseries=0,Rshunt=-1e3):
    """ Calculate the impedance over the frequency range
    
    Parameters
    ----------
    data : dataFrame
        Pandas dataFrame containing the time, voltage, current density and numerical error in the current density of the tj_file
    del_V : float
        Voltage step
    isToPlot : list
        List of array indices that will be used in the plotting

    Returns
    -------
    np.array
        Array of frequencies
    np.array
        Array of the real component of impedance
    np.array
        Array of the imaginary component of impedance
    np.array
        Array of complex error
    np.array
        Array of capacitance	
    np.array
        Array of conductance
    np.array
        Array of error in capacitance	
    np.array
        Array of error in conductance
    """
    # init the arrays for the impedance and its error:
    numFreqPoints = len(isToPlot)
    # V = np.empty(numFreqPoints)
    ReZ = np.empty(numFreqPoints)
    ImZ = np.empty(numFreqPoints)
    Z = [1 + 1J] * numFreqPoints
    errZ = [1 + 1J] * numFreqPoints
    C = np.empty(numFreqPoints)
    G = np.empty(numFreqPoints)
    errC = np.empty(numFreqPoints)
    errG = np.empty(numFreqPoints)


    # We need to know the series and shunt resistance to calculate the impedance correctly later on
    # dev_val,lyrs_dum = load_device_parameters(session_path, zimt_device_parameters)
    # for i in dev_val[zimt_device_parameters]:
    #     if i[0] == 'Contacts':
    #         contacts = i
    #         break

    # for i in contacts[1:]:
    #     if i[1] == 'R_series':
    #         Rseries = float(i[2])
    #     elif i[1] == 'R_shunt':
    #         Rshunt = float(i[2])

    for i in range(numFreqPoints):
        imax=isToPlot[i]
        if i == 0:
            data_dum = data.loc[0:imax]
        else:
            data_dum = data.loc[isToPlot[i-1]+1:imax]
        
        freq, Z[i], errZ[i] = calc_impedance_limit_time(np.asarray(data_dum['Jext']), np.asarray(data_dum['errJ']), np.asarray(data_dum['t']), del_V, imax)

        if Rseries > 0 and Rshunt > 0:
            # Correct the impedance for the series resistance
            Z[i] = Rseries + 1/(1/Z[i] + 1/Rshunt)
        elif Rseries > 0 and Rshunt < 0:
            # Correct the impedance for the series resistance
            Z[i] = Rseries + Z[i]
        elif Rshunt > 0 and Rseries <= 0:
            # Correct the impedance for the shunt resistance
            invZ = 1/Z[i]
            Z[i] = 1/(invZ + 1/Rshunt)

        invZ = 1/Z[i]
        
        # we are only interested in the absolute value of the real and imag components:
        ReZ[i] = Z[i].real
        ImZ[i] = Z[i].imag
        C[i] = 1/(2*math.pi*freq)*invZ.imag
        G[i] = invZ.real

        errC[i] = abs(1/(2*math.pi*freq)*(invZ.imag**2)*errZ[i].real)
        errG[i] = abs((invZ.real**2)*errZ[i].imag)
    
    return ReZ, ImZ, errZ, C, G, errC, errG

def store_capacitance_data(session_path, V, cap, errC, output_file='CapVol.dat'):
    """ Save the capacitance & its error in one file called CapVol.dat
    
    Parameters
    ----------
    session_path : string
        working directory for zimt
    V : np.array
        Array of frequencies
    cap : np.array
        Array of the capacitance
    errC : np.array
        Array of the capacitance error
    output_file : string
        Filename where the capacitance data is stored
    """

    with open(os.path.join(session_path,output_file), 'w') as file:
        file.write('V C errC' + '\n')
        for i in range(len(V)):
            file.write(f'{V[i]:.6e} {cap[i]:.6e} {errC[i]:.6e}' + '\n')

def get_capacitance(data, freq, V_0, V_max, del_V, V_step, session_path, zimt_device_parameters, output_file):
    """Calculate the capacitance from the simulation result

    Parameters
    ----------
    data : DataFrame
        DataFrame with the simulation results (tj.dat) file
    freq : float
        Frequency at which the capacitance-voltage measurement is performed
    V_0 : float
        Initial voltage
    V_max : float
        Maximum voltage
    del_V : float
        Voltage step that is applied directly after t=0
    V_step : float
        Voltage difference, determines at which voltages the capacitance is determined
    session_path : string
        working directory for zimt
    output_file : string
        Filename where the capacitance data is stored

    Returns
    -------
    integer,string
        returns -1 (failed) or 1 (success), including a message
    """  
    # get indexes of the time points where the voltage changes
    isToPlot = data.index[data['t'] == 0]
    # do -1 to get the last index
    isToPlot = isToPlot - 1
    # remove the first index, because we start at t=0
    isToPlot = isToPlot[1:]
    # add the last index to the end
    isToPlot = np.append(isToPlot, len(data['t'])-1)
    V = np.linspace(V_0, V_max, num=math.ceil((V_max-V_0)/V_step)+1, endpoint=True)

    ReZ, ImZ, errZ, C, G, errC, errG = calc_impedance_CV(data, del_V, isToPlot,session_path, zimt_device_parameters)

    # Write the capacitance results to a file
    store_capacitance_data(session_path, V, C, errC, output_file)

    # If the capacitance is calculated, message a success
    msg = 'Success'
    retval = 0
    
    return retval, msg

def cap_plot(session_path, output_file, xscale='linear', yscale='linear', plot_type = plt.errorbar):
    """ Plot the capacitance against voltage

    Parameters
    ----------
    session_path : string
        working directory for zimt
    output_file : string
        Filename where the capacitance data is stored
    xscale : string
        Scale of the x-axis. E.g linear or log
    yscale : string
        Scale of the y-axis. E.g linear or log
    plot_type : matplotlib.pyplot
        Type of plot to display
    """
    # Read the data from CapVol-file
    data = pd.read_csv(os.path.join(session_path,output_file), sep=r'\s+')

    # Define the plot parameters
    fig, ax = plt.subplots()
    pars = {'C' : 'Capacitance' }
    par_x = 'V'
    xlabel = 'Voltage [V]'
    ylabel = 'Capacitance [Fm$^{-2}$]'
    title = 'Capacitance-Voltage'
    
    # Plot with or without errorbars
    if plot_type == plt.errorbar:
        ax = utils_plot.plot_result(data, pars, list(pars.keys()), par_x, xlabel, ylabel, xscale, yscale, title, ax, plot_type, 
                                            [], data['errC'], legend=False)
    else:
        ax = utils_plot.plot_result(data, pars, list(pars.keys()), par_x, xlabel, ylabel, xscale, yscale, title, ax, plot_type, legend=False)

    plt.show()

def MottSchottky_plot(session_path, output_file, xscale='linear', yscale='linear', plot_type = plt.errorbar):
    """ Plot the 1/capacitance^2 against voltage

    Parameters
    ----------
    session_path : string
        working directory for zimt
    output_file : string
        Filename where the capacitance data is stored
    xscale : string
        Scale of the x-axis. E.g linear or log
    yscale : string
        Scale of the y-axis. E.g linear or log
    plot_type : matplotlib.pyplot
        Type of plot to display
    """
    # Read the data from CapVol-file
    data = pd.read_csv(os.path.join(session_path,output_file), sep=r'\s+')

    # Define the plot parameters
    fig, ax = plt.subplots()
    pars = {'1/C2' : 'Capacitance' }
    par_x = 'V'
    xlabel = 'Voltage [V]'
    ylabel = '1/Capacitance$^2$ [Fm$^{-2}$]$^{-2}$'
    title = 'Mott-Schottky'
    
    data['1/C2'] = 1/(data['C'])**2

    # Plot with or without errorbars
    if plot_type == plt.errorbar:
        ax = utils_plot.plot_result(data, pars, list(pars.keys()), par_x, xlabel, ylabel, xscale, yscale, title, ax, plot_type, 
                                            [], data['errC'], legend=False)
    else:
        ax = utils_plot.plot_result(data, pars, list(pars.keys()), par_x, xlabel, ylabel, xscale, yscale, title, ax, plot_type, legend=False)

    plt.show()

def plot_capacitance(session_path, output_file='CapVol.dat'):
    """Make a plot of the capacitance against voltage and the Mott-Schottky plot

    Parameters
    ----------
    session_path : string
        working directory for zimt
    """

    cap_plot(session_path, output_file)

    MottSchottky_plot(session_path, output_file)

def run_CV_simu(zimt_device_parameters, session_path, freq, V_min, V_max, V_step, G_frac=1, del_V=0.01,  run_mode=False,tVG_name='tVG.txt',  output_file = 'CapVol.dat', tj_name = 'tj.dat', varFile = 'none', ini_timeFactor=1e-3, timeFactor=1.02,**kwargs):
    """Create a tVG file and run ZimT with capacitance device parameters

    Parameters
    ----------
    zimt_device_parameters : string
        Name of the zimt device parameters file
    session_path : string, optional
        working directory for zimt
    freq : float
        Frequency at which the capacitance-voltage measurement is performed
    V_min : float
        Initial voltage
    V_max : float
        Maximum voltage
    V_step : float
        Voltage difference, determines at which voltages the capacitance is determined
    G_frac : float, optional
        Fractional light intensity, by default 1
    del_V : float, optional
        Voltage step that is applied directly after t=0, by default 0.01
    run_mode : bool, optional
        Indicate whether the script is in 'web' mode (True) or standalone mode (False). Used to control the console output, by default False  
    tVG_name : string, optional
        Name of the tVG file, by default tVG.txt
    output_file : string, optional
        Name of the file where the capacitance data is stored, by default CapVol.dat
    tj_name : string, optional
        Name of the tj file where the capacitance data is stored, by default tj.dat
    varFile : string, optional
        Name of the var file, by default 'none'
    ini_timeFactor : float, optional
        Constant defining the size of the initial timestep, by default 1e-3
    timeFactor : float, optional
        Exponential increase of the timestep, to reduce the amount of timepoints necessary. Use values close to 1., by default 1.02

    Returns
    -------
    CompletedProcess
        Output object of with returncode and console output of the simulation
    string
        Return message to display on the UI, for both success and failed
    """
    verbose = kwargs.get('verbose', False) # Check if the user wants to see the console output
    UUID = kwargs.get('UUID', '') # Check if the user wants to add a UUID to the tj file name
    cmd_pars = kwargs.get('cmd_pars', None) # Check if the user wants to add additional command line parameters
    # Check if the user wants to force the use of thread safe mode, necessary for Windows with parallel simulations
    if os.name == 'nt':
        threadsafe = kwargs.get('threadsafe', True) # Check if the user wants to force the use of threads instead of processes
    else:
        threadsafe = kwargs.get('threadsafe', False) # Check if the user wants to force the use of threads instead of processes
    
    # Update the file names with the UUID
    if UUID != '':
        dum_str = f'_{UUID}'
    else:
        dum_str = ''

    turnoff_autoTidy = kwargs.get('turnoff_autoTidy', None) # Check if the user wants to turn off the autoTidy function in SIMsalabim
    if turnoff_autoTidy is None: 
        if not threadsafe:
            turnoff_autoTidy = True
        else:
            turnoff_autoTidy = False

    # Update the filenames with the UUID
    tj_name = os.path.join(session_path, tj_name)
    output_file = os.path.join(session_path, output_file)
    tVG_name = os.path.join(session_path, tVG_name)
    if UUID != '':
        tj_file_name_base, tj_file_name_ext = os.path.splitext(tj_name)
        tj_name = tj_file_name_base + dum_str + tj_file_name_ext 
        tVG_name_base, tVG_name_ext = os.path.splitext(tVG_name)
        tVG_name = tVG_name_base + dum_str + tVG_name_ext
        output_file_base, output_file_ext = os.path.splitext(output_file)
        output_file = output_file_base + dum_str + output_file_ext
        if varFile != 'none':
            var_file_base, var_file_ext = os.path.splitext(varFile)
            varFile = var_file_base + dum_str + var_file_ext
            varFile = os.path.join(session_path,varFile)
    # varFile = 'none' # we don't use a var file for this simulation

    # The simulations with Rseries and Rshunt often do not converge, so we first run a steady state simulation to get the internal voltage and then run the impedance simulation with Rseries = 0 and Rshunt = -Rshunt. We will correct the impedance afterwards. This is a workaround to improve the convergence of the impedance simulation that should remain accurate to estimate the impedance.
    #default values for Rseries and Rshunt
    Rseries = 0
    Rshunt = -1e3

    # Do the steady state simulation to calculate the internal voltage in case of series resistance
    # Create tVG
    result, message = create_tVG_SS_CV(V_min, V_max, V_step, G_frac, tVG_name, session_path)

    # Get the device parameters and Rseries and Rshunt
    dev_val,layers_dum = load_device_parameters(session_path, zimt_device_parameters)
    for i in dev_val[zimt_device_parameters]:
        if i[0] == 'Contacts':
            contacts = i
            break

    for i in contacts[1:]:
        if i[1] == 'R_series':
            Rseries = float(i[2])
        elif i[1] == 'R_shunt':
            Rshunt = float(i[2])

    # Check if R_series and R_shunt are defined in cmd_pars
    idx_Rseries, idx_Rshunt = None, None
    if cmd_pars is not None:
        for idx, i in enumerate(cmd_pars):
            if i['par'] == 'R_series':
                Rseries = float(i['val'])
                idx_Rseries = idx
            elif i['par'] == 'R_shunt':
                Rshunt = float(i['val'])
                idx_Rshunt = idx

    if Rseries > 0:
        # Check if tVG file is created
        if result == 0:
            # In order for zimt to converge, set absolute tolerance of Poisson solver small enough
            tolPois = 10**(math.floor(math.log10(abs(del_V)))-4)

            # Define mandatory options for ZimT to run well with impedance:
            CV_SS_args = [{'par':'dev_par_file','val':zimt_device_parameters},
                                {'par':'tVGFile','val':tVG_name},
                                {'par':'tolPois','val':str(tolPois)},
                                {'par':'limitDigits','val':'0'},
                                {'par':'currDiffInt','val':'2'},
                                {'par':'tJFile','val':tj_name},
                                {'par':'varFile','val':varFile},
                                {'par':'logFile','val':'log'+dum_str+'.txt'}
                                ]
            if turnoff_autoTidy:
                CV_SS_args.append({'par':'autoTidy','val':'0'})

            if cmd_pars is not None:
                CV_SS_args = update_cmd_pars(CV_SS_args, cmd_pars)
            
            if threadsafe:
                result, message = utils_gen.run_simulation_filesafe('zimt', CV_SS_args, session_path, run_mode, verbose=verbose)
            else:
                result, message = utils_gen.run_simulation('zimt', CV_SS_args, session_path, run_mode, verbose=verbose)
            
            if result == 0 or result == 95:
                data = read_tj_file(session_path, tj_file_name=tj_name)

                Vext = np.asarray(data['Vext'])
                Jext = np.asarray(data['Jext'])
                Vint = Vext - Jext*Rseries
            else:
                return result, message
        else:
            return result, message

        # Create tVG
        result, message = create_tVG_CV_Rseries(Vint, del_V, G_frac, tVG_name, session_path, freq, ini_timeFactor, timeFactor)

    else:            
        # Create tVG
        result, message = create_tVG_CV(V_min, V_max, del_V, V_step, G_frac, tVG_name, session_path, freq, ini_timeFactor, timeFactor)

    # remove the Rseries and Rshunt from cmd_pars
    if idx_Rseries is not None and idx_Rshunt is not None:
        if idx_Rseries < idx_Rshunt:
            cmd_pars.pop(idx_Rshunt)
            cmd_pars.pop(idx_Rseries)
        else:
            cmd_pars.pop(idx_Rseries)
            cmd_pars.pop(idx_Rshunt)
    elif idx_Rseries is not None and idx_Rshunt is None:
        cmd_pars.pop(idx_Rseries)
    elif idx_Rshunt is not None and idx_Rseries is None:
        cmd_pars.pop(idx_Rshunt)

    # Check if tVG file is created
    if result == 0:
        # In order for zimt to converge, set absolute tolerance of Poisson solver small enough
        tolPois = 10**(math.floor(math.log10(abs(del_V)))-4)

        # Define mandatory options for ZimT to run well with CV:
        CV_args = [{'par':'dev_par_file','val':zimt_device_parameters},
                        {'par':'tVGFile','val':tVG_name},
                        {'par':'tolPois','val':str(tolPois)},
                        {'par':'limitDigits','val':'0'},
                        {'par':'currDiffInt','val':'2'},
                        {'par':'tJFile','val':tj_name},
                        {'par':'varFile','val':varFile},
                        {'par':'logFile','val':'log'+dum_str+'.txt'},
                        # We remove Rseries and Rshunt as the simulation is either to converge that way, we will correct the impedance afterwards
                        {'par':'R_series','val':str(0)},
                        {'par':'R_shunt','val':str(-abs(Rshunt))}]
        
        if turnoff_autoTidy:
            CV_args.append({'par':'autoTidy','val':'0'})
            
        if cmd_pars is not None:
                CV_args = update_cmd_pars(CV_args, cmd_pars)

        if threadsafe:
            result, message = utils_gen.run_simulation_filesafe('zimt', CV_args, session_path, run_mode, verbose=verbose)
        else:
            result, message = utils_gen.run_simulation('zimt', CV_args, session_path, run_mode, verbose=verbose)

        if result == 0 or result == 95:
            data = read_tj_file(session_path, tj_file_name=tj_name)

            result, message = get_capacitance(data, freq, V_min, V_max, del_V, V_step, session_path, zimt_device_parameters, output_file)
            return result, message

        else:
            return result, message
        
    return result, message

######### Running the function as a standalone script #############################################
if __name__ == "__main__":
    ## Manual CV input parameters. These are overwritten by command line arguments if provided
    freq = 1e4
    V_min = 0.5
    V_max = 1.0
    del_V = 10e-3
    V_step = 0.1    
    G_frac = 0

    # Define folder and file paths
    session_path = os.path.join('../../','SIMsalabim','ZimT')

    zimt_device_parameters = 'simulation_setup.txt'

    tVG_name = 'tVG.txt'
    tj_name = 'tj.dat'
    output_name = 'freqZ.dat'

    # UUID = str(uuid.uuid4()) # Add a UUID to the simulation
    UUID = ''

    # Not user input
    ini_timeFactor = 1e-3 # Initial timestep factor, org 1e-3
    timeFactor = 1.02 # Increase in timestep every step to reduce the amount of datapoints necessary, use value close to 1 as this is best! Org 1.02
    run_mode = False # If False, show verbose output in console

    ############## Command line arguments  ##############
    ## Notes
    ## - The command line arguments are optional and can be provided in any order
    ## - Each command line argument must be provided in the format -par_name value
    ## - Possible arguments include all SIMsalabim parameters and CV specific parameters as listed before
    ## - Special arguments
    ##   - sp : string
    ##     - The session path, i.e. the working directory for the simulation
    ##   - simsetup : string
    ##     - The name of the zimt simulation setup parameters file
    ##   - UUID : string
    ##     - An UUID to add to the simulation (output)

    cmd_pars_dict = {}
    cmd_pars = []

    # Check if any arguments are provided.
    if len(sys.argv) >= 2:
        # Each arguments should be in the format -par val, i.e. in pairs of two. Skip the first argument as this is the script name
        if not len(sys.argv[1:]) % 2 == 0:
            print('Error in command line parameters. Please provide arguments in the format -par_name value -par_name value ...')
            sys.exit(1)
        else:
            input_list = sys.argv[1:]
            # Loop over the input list and put pairs into a dictionary with the first argument as key and the second argument as value
            for i in range(0,len(input_list),2):
                # Check if the key already exists, if not, add it to the cmd_pars_dict
                if str(input_list[i][1:]) in cmd_pars_dict:
                    print(f'Duplicate parameter found in the command line parameters: {str(input_list[i][1:])}')
                    sys.exit(1)
                else:
                    cmd_pars_dict[str(input_list[i][1:])] = str(input_list[i+1])

    # Check and process specific keys/arguments that are not native SIMsalabim arguments
    # Handle the session_path/sp argument separately, as the other parameters depend on this
    if 'sp' in cmd_pars_dict:
        session_path = cmd_pars_dict['sp']
        # remove from cmd_pars
        cmd_pars_dict.pop('sp')

    # Define mappings for keys and variables
    key_action_map = {
        'simsetup': lambda val: {'zimt_device_parameters': val},
        'freq': lambda val: {'freq': float(val)},
        'V_min': lambda val: {'V_min': float(val)},
        'V_max': lambda val: {'V_max': float(val)},
        'del_V': lambda val: {'del_V': float(val)},
        'V_step': lambda val: {'V_step': float(val)},
        'G_frac': lambda val: {'G_frac': float(val)},
        'tVG_name': lambda val: {'tVG_name': val},
        'tj_name': lambda val: {'tj_name': val},
        'out_name': lambda val: {'output_name': val},
        'UUID': lambda val: {'UUID': val},    
    }

    for key in list(cmd_pars_dict.keys()):  # Use list to avoid modifying the dictionary while iterating
        if key in key_action_map:
            # Apply the corresponding action
            result = key_action_map[key](cmd_pars_dict[key])
            globals().update(result)  # Dynamically update global variables
            cmd_pars_dict.pop(key)

    # Handle remaining keys in `cmd_pars_dict` and add them to the cmd_pars list
    cmd_pars.extend({'par': key, 'val': value} for key, value in cmd_pars_dict.items())

    # Run Capacitance-Voltage   
    result, message = run_CV_simu(zimt_device_parameters, session_path, freq, V_min, V_max, V_step, G_frac, del_V,run_mode=False, tVG_name=tVG_name, 
                                    output_file = output_name, tj_name = tj_name, ini_timeFactor=ini_timeFactor, timeFactor=timeFactor, cmd_pars=cmd_pars, UUID=UUID)

    # Make the capacitance-voltage plot
    if result == 0 or result == 95:
        plot_capacitance(session_path, os.path.basename(output_name))
    else:
        print(message)
        sys.exit(1)


    



"""Perform EQE simulations"""

######### Package Imports #########################################################################

import os, uuid, sys
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy import constants
# import pySIMsalabim
## Import pySIMsalabim, if not successful, add the parent directory to the system path
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.utils import general as utils_gen
from pySIMsalabim.utils.parallel_sim import *
from pySIMsalabim.utils.utils import update_cmd_pars

######### Constants #################################################################################

h=constants.h
c=constants.c
q=constants.e

######### Functions #################################################################################

def EQE_create_spectrum_files(lambda_array, p, session_path, spectrum_path, tmp_spectrum_path):
    """Create all the spectrum files with monochromatic peaks at the specified wavelength, 
    with a fixed height set by the number of added photons p. The files are stored in a temporary folder: tmp_spectrum.

    Parameters
    ----------
    lambda_array : array
        Array with the wavelengths at which the monochromatic peaks will be created.
    p : float
        Number of photons that are added to the irradiance.
    session_path : string
        Path to the session folder where the files must be created.
    spectrum_path : string
        Path to the original spectrum file.
    tmp_spectrum_path : string
        Path to the temporary folder where the modified spectrum files will be stored.
    """
    for i in lambda_array:
        # Reset the spectrum data on every iteration to avoid accumulation of the added photons at each wavelength
        org_spectrum_data = pd.read_csv(os.path.join(session_path,spectrum_path), sep=r'\s+')

        # Get the filename of the spectrum without any possible path
        spectrum_filename = os.path.basename(spectrum_path)

        #find row with value closest to i and multiply the value in the irradiance column with 10
        row = org_spectrum_data.iloc[(org_spectrum_data['lambda']-i).abs().argsort()[:1]]
        org_spectrum_data.loc[row.index, 'I'] += (p*((h*c)/i)/1e-9) #2% of absorbed photons of Si
        org_spectrum_data.to_csv(os.path.join(tmp_spectrum_path,f'{int(i*1e9)}nm_{spectrum_filename}'), sep=' ', index=False, float_format='%.3e')

def get_CurrDens(JV_file, session_path):
    """ Get the current density  and its from the JV_file as stored in the first row.
    Parameters
    ----------
    JV_file : str, optional
        Name of the file where the JV data is stored. Must be unique for each simulation, by default 'JV.dat'
    session_path : string
        Path to the session folder where the simulation will run. 

    Returns
    -------
    float, float
        Short-circuit current and its error.
    """     
    data_JV = pd.read_csv(os.path.join(session_path,JV_file), sep=r'\s+')
    J0=data_JV['Jext'][0]
    J0_err=data_JV['errJ'][0]

    # Remove the JV file as it is not needed anymore
    os.remove(os.path.join(session_path,JV_file))

    return J0, J0_err

def calc_EQE(Jext,Jext_err,J0_single, J0_err_single, lambda_array,p):
    """Calculate the EQE values for the monochromatic peaks at each wavelength. Based on the change in the short-circuit current and the number of added photons.

    Parameters
    ----------
    Jext : array
        Short-circuit current density for each monochromatic peak.
    Jext_err : array
        Error in the short-circuit current density for each monochromatic peak.
    J0_single : float
        Short-circuit current density for the normal spectrum.
    J0_err_single : float
        Error in the short-circuit current density for the normal spectrum.
    lambda_array : array
        Array with the wavelengths at which the monochromatic peaks were created.
    p : float
        Number of photons that are added to the irradiance.

    Returns
    -------
    array, array, array, array, array
        Arrays with the change in short-circuit current density, error in the change in short-circuit currentdensity,
        monochromatic intensity, EQE values and error in the EQE values.
    """
    # Calculate the change in short-circuit current density and its error
    deltaJ = [abs(x - J0_single) for x in Jext]
    deltaJerr = [x + J0_err_single for x in Jext_err]

    # Calculate the increase in intensity/photon flux/energy of the monochromatic peaks
    I_diff = [((p * h * c) / x) / 1e-9 for x in lambda_array]
    area = 1e-9 * np.array(I_diff)
    E_photon = np.array([(h * c) / x for x in lambda_array])

    # Calculate the EQE values and their errors
    EQE_val=deltaJ/((q*area)/E_photon)
    EQE_err=np.multiply(deltaJerr,((q*area)/E_photon))

    return deltaJ, deltaJerr, I_diff, EQE_val, EQE_err

def run_EQE(simss_device_parameters, session_path, spectrum, lambda_min, lambda_max, lambda_step, Vext, output_file = 'EQE.dat', JV_file_name = 'JV.dat', varFile = 'none',remove_dirs = True, parallel = False, max_jobs = max(1,os.cpu_count()-1), run_mode = True, **kwargs):
    """Run the EQE calculation for a given spectrum and external voltage, and save the results in a file.

    Parameters
    ----------
    simss_device_parameters : string
        Name of the device parameters file.
    session_path : string
        Path to the session folder where the simulation will run.
    spectrum : string
        Path to the original spectrum file.
    lambda_min : float
        Minimum wavelength for the spectrum.
    lambda_max : float
        Maximum wavelength for the spectrum.
    lambda_step : float
        Step size for the wavelength.
    Vext : float
        External voltage at which the simulation will run and the EQE must be calculated.
    output_file : string, optional
        Name of the file where the results will be stored. The file will be stored in the session folder, by default 'EQE.dat'
    JV_file_name : string, optional
        Name of the JV file. Must be unique for each simulation, by default 'JV.dat'
    varFile : string, optional
        Name of the var file, by default 'none'
    remove_dirs : bool, optional
        Remove the temporary directories, by default True
    parallel : bool, optional
        Run the simulations in parallel, by default False
    max_jobs : int, optional
        Maximum number of parallel jobs, by default max(1,os.cpu_count()-1)
    run_mode : bool, optional
        indicate whether the script is in 'web' mode (True) or standalone mode (False). Used to control the console output, by default True
    **kwargs : dict
        Additional arguments to be passed to the function.

    Returns
    -------
    int
        0 if the function runs successfully.
    """
    verbose = kwargs.get('verbose', False) # Check if the user wants to print messages to the console
    UUID = kwargs.get('UUID', '') # Check if the user wants to add a UUID to the JV file name
    cmd_pars = kwargs.get('cmd_pars', None) # Check if the user wants to add additional command line parameters to the simulation
    force_multithreading = kwargs.get('force_multithreading', False) # Check if the user wants to force multithreading instead of using GNU parallel 
    if os.name == 'nt':  
        threadsafe = kwargs.get('threadsafe', True) # Check if the user wants to force the use of threads instead of processes
    else:
        threadsafe = kwargs.get('threadsafe', False) # Check if the user wants to force the use of threads instead of processes

    turnoff_autoTidy = kwargs.get('turnoff_autoTidy', None) # Check if the user wants to turn off the autoTidy function in SIMsalabim
    if turnoff_autoTidy is None: 
        if not threadsafe:
            turnoff_autoTidy = True
        else:
            turnoff_autoTidy = False

    # Update the JV file name with the UUID
    if UUID != '':
        dum_str = f'_{UUID}'
    else:
        dum_str = ''

    
    JV_file_name = os.path.join(session_path, JV_file_name)
    output_file = os.path.join(session_path, output_file)
    if UUID != '':
        JV_file_name_base, JV_file_name_ext = os.path.splitext(JV_file_name)
        JV_file_name = JV_file_name_base + dum_str + JV_file_name_ext
        output_file_base, output_file_ext = os.path.splitext(output_file)
        output_file = output_file_base + dum_str + output_file_ext
        if varFile != 'none':
            varFile = os.path.join(session_path,varFile)
            var_file_base, var_file_ext = os.path.splitext(varFile)
            varFile = var_file_base + dum_str + var_file_ext
    # varFile = 'none' # we don't use a var file for this simulation
    
    msg_list = [] # Init the returnmessage list
    p=0.03*1e21 #number of photons that are added to the irradiance. in this example it is 3% of the number of photons absorbed by a Silicon solar cell in m-2
    
    # Create a tmp folder to store the modified AM1.5G spectra
    # rnd_ID = str(uuid.uuid4())
    tmp_spectrum_path = os.path.join(session_path,'tmp_spectrum'+UUID)
    if not os.path.exists(tmp_spectrum_path):
        os.makedirs(tmp_spectrum_path)

    # Create wavelength array. If the last point in the array is outside the specified range, remove it
    lambda_array_init = np.arange(lambda_min,lambda_max+lambda_step,lambda_step)
    if lambda_array_init[-1] > lambda_max:
        lambda_array = np.delete(lambda_array_init,-1)
    else:
        lambda_array = lambda_array_init

    # convert to m
    lambda_array = lambda_array*1e-9

    EQE_create_spectrum_files(lambda_array, p, session_path, spectrum, tmp_spectrum_path)

    #runs for no monochromatic peak (normal spectrum) and obtains J0 and its err
    # Prepare the arguments for the simulation
    EQE_args = [{'par':'dev_par_file','val':simss_device_parameters},
                    {'par':'outputRatio','val':str(0)}, # we don't care about the var file here
                    {'par':'Vmin','val':str(Vext)},
                    {'par':'Vmax','val':str(Vext)},
                    {'par':'spectrum','val':spectrum},
                    {'par':'JVFile','val':os.path.join(session_path,JV_file_name)}, # need to add the session path to the JV file name to make sure it is stored in the session folder
                    # below are unecessary output files but we asign them to avoid errors
                    {'par':'logFile','val':'log'+dum_str+'.txt'},
                    {'par':'varFile','val':varFile},
                    {'par':'scParsFile','val':'scPars'+dum_str+'.txt'}
                    ]
    if turnoff_autoTidy:
        EQE_args.append({'par':'autoTidy','val':'0'})

    if cmd_pars is not None:
        EQE_args = update_cmd_pars(EQE_args, cmd_pars)
    
    if threadsafe:
        result, message = utils_gen.run_simulation_filesafe('simss', EQE_args, session_path, run_mode,verbose=verbose)
    else:
        result, message = utils_gen.run_simulation('simss', EQE_args, session_path, run_mode,verbose=verbose)
    # result, message = run_sim_EQE(simss_device_parameters, session_path, spectrum, Vext, JV_file_name, run_mode)
    
    # If the simulation fails, stop running the script and exit
    if not result == 0:
        if result== 95:
            message = f'SIMsalabim raised an error with errorcode {result}, simulation did not converge.'
        msg_list.append(message)
        return result, msg_list
    
    # Get the current density Jext and its error
    J0_single, J0_err_single = get_CurrDens(JV_file_name, session_path)

    Jext,Jext_err = [],[]

    #obtains Jext and Jext error for a monochromatic peak at each wavelength        
    if parallel and len(lambda_array) > 1 : # Run the simulations in parallel
        JV_file_name_base, JV_file_name_ext = os.path.splitext(JV_file_name)
        log_file_name_base, log_file_name_ext = os.path.splitext('log'+dum_str+'.txt')
        scParsFile_name_base, scParsFile_name_ext = os.path.splitext('scPars'+dum_str+'.txt')
        
        EQE_args_list = []
        for i in lambda_array:
            JV_file_name_single = f'{JV_file_name_base}{dum_str}_{int(i*1e9)}nm{JV_file_name_ext}'
            if varFile != 'none':
                varFile_single = f'{var_file_base}+{dum_str}_{int(i*1e9)}nm{var_file_ext}'
                varFile_single = os.path.join(session_path,varFile_single)
            else:
                varFile_single = 'none'
            dum_args = [{'par':'dev_par_file','val':simss_device_parameters},
                        {'par':'outputRatio','val':str(0)}, # we don't care about the var file here
                        {'par':'Vmin','val':str(Vext)},
                        {'par':'Vmax','val':str(Vext)},
                        {'par':'spectrum','val':os.path.join(tmp_spectrum_path,f'{int(i*1e9)}nm_{os.path.basename(spectrum)}')},
                        {'par':'JVFile','val':os.path.join(session_path,JV_file_name_single)}, # makes sure that the JV file is unique for each simulation and is stored in the session folder
                        # below are unecessary output files but we asign them to avoid errors
                        {'par':'logFile','val':f'{log_file_name_base}_{int(i*1e9)}nm{log_file_name_ext}'},
                        {'par':'varFile','val':varFile_single},
                        {'par':'scParsFile','val':f'{scParsFile_name_base}_{int(i*1e9)}nm{scParsFile_name_ext}'}
                        ]
            if turnoff_autoTidy:
                dum_args.append({'par':'autoTidy','val':'0'}) # necessary to avoid autoTidy messing up the files in parallel mode

            if cmd_pars is not None:
                dum_args = update_cmd_pars(dum_args, cmd_pars)
            EQE_args_list.append(dum_args)
        
        results = run_simulation_parallel('simss', EQE_args_list, session_path, max_jobs, force_multithreading=force_multithreading,verbose=verbose)
        
        for i in lambda_array:
            JV_file_name_single = f'{JV_file_name_base}{dum_str}_{int(i*1e9)}nm{JV_file_name_ext}'
            J_single, Jerr_single = get_CurrDens(os.path.join(session_path,JV_file_name_single),session_path)
            Jext.append(J_single)
            Jext_err.append(Jerr_single)
            
    else:
        for i in lambda_array:
            JV_file_name_base, JV_file_name_ext = os.path.splitext(JV_file_name)
            JV_file_name_single = f'{JV_file_name_base}_{int(i*1e9)}nm{JV_file_name_ext}'
            log_file_name_base, log_file_name_ext = os.path.splitext('log'+dum_str+'.txt')
            scParsFile_name_base, scParsFile_name_ext = os.path.splitext('scPars'+dum_str+'.txt')
            if varFile != 'none':
                varFile_single = f'{var_file_base}_{int(i*1e9)}nm{var_file_ext}'
                varFile_single = os.path.join(session_path,varFile_single)
            else:
                varFile_single = 'none'
            
            EQE_args = [{'par':'dev_par_file','val':simss_device_parameters},
                        {'par':'outputRatio','val':str(0)}, # we don't care about the var file here
                        {'par':'Vmin','val':str(Vext)},
                        {'par':'Vmax','val':str(Vext)},
                        {'par':'spectrum','val':os.path.join(tmp_spectrum_path,f'{int(i*1e9)}nm_{os.path.basename(spectrum)}')},
                        {'par':'JVFile','val':os.path.join(session_path,JV_file_name_single)},#JV_file_name_single}, # makes sure that the JV file is unique for each simulation and is stored in the session folder
                        # below are unecessary output files but we asign them to avoid errors
                        {'par':'logFile','val':f'{log_file_name_base}_{int(i*1e9)}nm{log_file_name_ext}'},
                        {'par':'varFile','val':varFile_single},
                        {'par':'scParsFile','val':f'{scParsFile_name_base}_{int(i*1e9)}nm{scParsFile_name_ext}'}
                        ]  
            if turnoff_autoTidy:
                EQE_args.append({'par':'autoTidy','val':'0'})

            if cmd_pars is not None:
                EQE_args = update_cmd_pars(EQE_args, cmd_pars)

            if threadsafe:
                result, message = utils_gen.run_simulation_filesafe('simss', EQE_args, session_path, run_mode,verbose=verbose)
            else:
                result, message = utils_gen.run_simulation('simss', EQE_args, session_path, run_mode,verbose=verbose)

            if not result == 0:
                msg_list.append(message)

            J_single, Jerr_single = get_CurrDens(JV_file_name_single,session_path)
            
            Jext.append(J_single)
            Jext_err.append(Jerr_single)


    # Remove the tmp folder as it is not needed anymore
    if remove_dirs:
        for file in os.listdir(tmp_spectrum_path):
            os.remove(os.path.join(tmp_spectrum_path,file))
        os.rmdir(os.path.join(tmp_spectrum_path))

    # remove all files starting with 'log'+dum_str and 'scPars'+dum_str
    for file in os.listdir(session_path):
        if file.startswith('log'+dum_str) or file.startswith('scPars'+dum_str):
            os.remove(os.path.join(session_path,file))
            
    # Calculate EQE
    deltaJ, deltaJerr, I_diff, EQE_val, EQE_err = calc_EQE(Jext,Jext_err,J0_single, J0_err_single, lambda_array,p)

    # Save the results in a file
    fp = open(os.path.join(session_path,output_file), 'w')
    fp.write('lambda Jext Jerr deltaJ deltaJerr Imonopeak EQE EQEerr\n')
    for i in range(len(lambda_array)):
        fp.write(f'{lambda_array[i]:.3e} {Jext[i]:.3e} {Jext_err[i]:.3e} {deltaJ[i]:.3e} {deltaJerr[i]:.3e} {I_diff[i]:.3e} {EQE_val[i]:.3e} {EQE_err[i]:.3e}\n')
    fp.close()
    
    if len(msg_list) != 0:
        return 1, msg_list
    else:
        return 0, msg_list

if __name__ == '__main__':
     ## Manual EQE input parameters. These are overwritten by command line arguments if provided
    lambda_min = 280 # Lower limit of the wavelength range in nm
    lambda_max = 1000 # Upper limit of the wavelength range in nm
    lambda_step = 20 # Step size of the wavelength range in nm
    Vext = [0] # Array with applied voltages in V

    # Define folder and file paths
    session_path = os.path.join('../../','SIMsalabim','SimSS')

    simss_device_parameters = 'simulation_setup.txt'
    spectrum = os.path.join('..','Data','AM15G.txt')  # Path of the original spectrum file

    # UUID = str(uuid.uuid4()) # Add a UUID to the simulation
    UUID = ''

    ############## Command line arguments  ##############
    ## Notes
    ## - The command line arguments are optional and can be provided in any order
    ## - Each command line argument must be provided in the format -par_name value
    ## - Possible arguments include all SIMsalabim parameters and EQE specific parameters as listed before
    ## - Special arguments
    ##   - sp : string
    ##     - The session path, i.e. the working directory for the simulation
    ##   - simsetup : string
    ##     - The name of the simss simulation setup parameters file
    ##   - UUID : string
    ##     - An UUID to add to the simulation (output)

    cmd_pars_dict = {}
    cmd_pars = []

 # Check if any arguments are provided.
    if len(sys.argv) >= 2:
        # Each arguments should be in the format -par val, i.e. in pairs of two. Skip the first argument as this is the script name
        if not len(sys.argv[1:]) % 2 == 0:
            print('Error in command line parameters. Please provide arguments in the format -par_name value -par_name value ...')
            sys.exit(1)
        else:
            input_list = sys.argv[1:]
            # Loop over the input list and put pairs into a dictionary with the first argument as key and the second argument as value
            for i in range(0,len(input_list),2):
                # Check if the key already exists, if not, add it to the cmd_pars_dict
                if str(input_list[i][1:]) in cmd_pars_dict:
                    print(f'Duplicate parameter found in the command line parameters: {str(input_list[i][1:])}')
                    sys.exit(1)
                else:
                    cmd_pars_dict[str(input_list[i][1:])] = str(input_list[i+1])

    # Check and process specific keys/arguments that are not native SIMsalabim arguments
    # Handle the session_path/sp argument separately, as the other parameters depend on this
    if 'sp' in cmd_pars_dict:
        session_path = cmd_pars_dict['sp']
        # remove from cmd_pars
        cmd_pars_dict.pop('sp')

    # Define mappings for keys and variables
    key_action_map = {
        'simsetup': lambda val: {'simss_device_parameters': val},
        'lambda_min': lambda val: {'lambda_min': float(val)},
        'lambda_max': lambda val: {'lambda_max': float(val)},
        'lambda_step': lambda val: {'lambda_step': int(val)},
        'Vext': lambda val: {'Vext': val.split(',')}, # IMPORTANT: if multiple voltages are provided, they must be separated by a comma!!
        'spectrum': lambda val: {'spectrum': val},
        'UUID': lambda val: {'UUID': val},
    }

    for key in list(cmd_pars_dict.keys()):  # Use list to avoid modifying the dictionary while iterating
        if key in key_action_map:
            # Apply the corresponding action
            result = key_action_map[key](cmd_pars_dict[key])
            globals().update(result)  # Dynamically update global variables
            cmd_pars_dict.pop(key)

    # Handle remaining keys in `cmd_pars_dict` and add them to the cmd_pars list
    cmd_pars.extend({'par': key, 'val': value} for key, value in cmd_pars_dict.items())

    for i in range(len(Vext)):
        output_file = f'output_{Vext[i]}V.dat'
        retval = run_EQE(simss_device_parameters, session_path, spectrum, lambda_min, lambda_max, lambda_step, Vext[i], output_file, remove_dirs=True, run_mode=False, UUID=UUID)
    
    plt.figure()
    for i in range(len(Vext)):
        res = pd.read_csv(os.path.join(session_path,f'output_{Vext[i]}V.dat'), sep=r'\s+')
        plt.scatter(res['lambda']/1E-9, res['EQE'], label=f'V = {i}')
        plt.title('EQE for solar cell', fontsize = 16)
        plt.tick_params(axis='both',direction='in')
        plt.xlabel('Wavelength [nm]', fontsize = 16)
        plt.ylabel('EQE', fontsize = 16)
        plt.xlim(lambda_min, lambda_max)
        plt.ylim(0,1)
        plt.legend()

    plt.show()



""" Test the JV module of pySIMsalabim """

######### Package Imports #########################################################################
import os, sys, uuid 
import pandas as pd
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.experiments.JV_steady_state import *

######### Test Functions #########################################################################

def test_run_SS_JV():
    """ Test the run_SS_JV function """
    # Set the path to the simulation setup file
    if os.path.exists('SIMsalabim'):
        cwd = os.path.abspath('.')
    else:
        cwd = os.path.abspath('../..')
    simss_device_parameters = os.path.join(cwd, 'SIMsalabim','SimSS','simulation_setup.txt')
    session_path = os.path.join(cwd, 'SIMsalabim','SimSS')
    # Set the JV parameters
    Gfracs = [0.1,0.5,1.0]
    UUID = str(uuid.uuid4())
    # Run the JV simulation
    
    ret, mess = run_SS_JV(simss_device_parameters, session_path, JV_file_name = 'JV.dat', G_fracs = Gfracs, parallel = False, max_jobs = 3, run_mode = False, UUID=UUID, cmd_pars=[{'par': 'l2.L', 'val': '500e-9'}])
    print(ret,mess)
    # Check the output
    assert ret == 0, 'JV simulation failed'

def test_SS_JV_parallel():
    """ Test the SS_JV_parallel function """
    try:
        # check if SIMsalabim folder exists in the current directory
        if os.path.exists('SIMsalabim'):
            cwd = os.path.abspath('.')
        else:
            cwd = os.path.abspath('../..')
        simss_device_parameters = os.path.join(cwd, 'SIMsalabim','SimSS','simulation_setup.txt')
        session_path = os.path.join(cwd, 'SIMsalabim','SimSS')

        def run(G_frac,ID): # function to run the JV simulation in parallel
            cmd_pars = [{'par': 'l2.L', 'val': '500e-9'}]
            print('Running')
            res = run_SS_JV(simss_device_parameters, session_path, JV_file_name = 'JV.dat', G_fracs = [G_frac], parallel = False, max_jobs = 3, run_mode = False, UUID=ID, cmd_pars=cmd_pars)
            return res

        # Run the JV simulation in parallel
        Gfracs = [0.1,0.5,1.0]
        ID_list = [str(uuid.uuid4()) for i in range(len(Gfracs))]
        res = Parallel(n_jobs=3)(delayed(run)(Gfrac,ID) for Gfrac,ID in zip(Gfracs,ID_list))
        
        # print('Plotting')
        # plt.figure()
        # for Gfrac,UUID in zip(Gfracs,ID_list):
        #     data = pd.read_csv(os.path.join(session_path,f'JV_Gfrac_{Gfrac}_{UUID}.dat'), sep=r'\s+')
        #     plt.plot(data['Vext'],data['Jext'],label=f'Gfrac = {Gfrac}')
        # plt.xlabel('Vext [V]')
        # plt.ylabel('Current density [A/m^2]')
        # plt.legend()
        # plt.show()

        # sim.clean_all_output(session_path)
        # sim.clean_up_output('Gfracs',session_path)
        # sim.delete_folders('tmp',session_path)
        assert True
    except Exception as e:
        print(e)
        assert False, 'JV parallel simulation failed'


if __name__ == '__main__':
    test_run_SS_JV()
    test_SS_JV_parallel()

    cwd = os.path.abspath('../..')
    session_path = os.path.join(cwd, 'SIMsalabim','SimSS')
    sim.clean_all_output(session_path)
    sim.clean_up_output('Gfracs',session_path)
    sim.delete_folders('tmp',session_path)

    print('All JV tests passed')